import unittest
import os
import sys
sys.path.insert(0, '../')
import starseqr_utils as su

path = os.path.dirname(__file__)
if path != '':
    os.chdir(path)


class DiversityTestCase(unittest.TestCase):
    '''Tests cross homology'''

    def test_overhang_v1(self):
        """test homology paired.fastq TUBA1B--TUBA1A test case"""
        res1 = su.overhang_diversity.get_diversity('chr12:49521770:-:chr12:49578857:-:6:0')
        assert(res1 == 2)

    def test_crosshom_v2(self):
        """test diversity with no reads in fastq test case"""
        res1 = su.overhang_diversity.get_diversity('chr2:131389025:+:chr2:131996848:+:0:4')
        assert(res1 == 0)

    def test_crosshom_v3(self):
        """test homology with lots of reads in fastq test case"""
        res1 = su.overhang_diversity.get_diversity('chr1:45268528:+:chr4:107152937:-:0:2')
        assert(res1 == 6)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8

from __future__ import (absolute_import, division, print_function)
import logging
import ssw
import numpy as np
import starseqr_utils as su

logger = logging.getLogger('STAR-SEQR')


def get_ssw_scores(reads_fq, trxleft_fa, trxright_fa, mm_rate=2, mm_pct=.05):
    '''get sw score for left and right gene'''
    aligner = ssw.Aligner(gap_open=12, gap_extend=4)
    rfq_gen = su.common.FastqParser(reads_fq)
    matches = []
    for rfq in rfq_gen:
        l_max = 0
        r_max = 0
        mod_len = min(rfq.seq_len - mm_rate , rfq.seq_len - rfq.seq_len * mm_pct) # control error rates across a diversity of read len
        mod_len = mod_len
        trxl_gen = su.common.fasta_iter(trxleft_fa)
        trxr_gen = su.common.fasta_iter(trxright_fa)
        for trxl_id, trxl_seq in trxl_gen:
            rfql_align = aligner.align(reference=rfq.sequence, query=trxl_seq)
            l_max = max(rfql_align.score, l_max)
        for trxr_id, trxr_seq in trxr_gen:
            rfqr_align = aligner.align(reference=rfq.sequence, query=trxr_seq)
            r_max = max(rfqr_align.score, r_max)
        read_norm = min(l_max, r_max) / mod_len # pct identity matching
        matches.append(read_norm / 2)
    if len(matches) > 0:
        return(float("{0:.3f}".format(np.sum(i > .8 for i in matches) / len(matches))))
        # return(float("{0:.3f}".format(np.sum(i > 32 for i in matches) / len(matches)))) # switch to proportion of reads > 32 (16bp homology) #TODO profile
        # return(int(np.percentile(matches, 75)))  # UQ75 of ssw from cross-mapping
    else:
        return(0)


def get_cross_homology(jxn):
    paired_res = None
    overhang_res = None
    clean_jxn = su.common.safe_jxn(jxn)
    jxn_dir = 'support' + '/' + clean_jxn + '/'

    trans_left_fa = jxn_dir + 'transcripts_left.fa'
    trans_right_fa = jxn_dir + 'transcripts_right.fa'
    paired_fq = jxn_dir + 'span.fastq' # this file contains overhang/junction reads.
    overhang_fq = jxn_dir + 'overhang.fastq'
    paired_res = get_ssw_scores(paired_fq, trans_left_fa, trans_right_fa)
    overhang_res = get_ssw_scores(overhang_fq, trans_left_fa, trans_right_fa)
    return paired_res, overhang_res



#!/usr/bin/env python
# encoding: utf-8

from __future__ import (absolute_import, division, print_function)
import six
import re
import logging
from collections import defaultdict
import pandas as pd


logger = logging.getLogger("STAR-SEQR")

try:
    maketrans = ''.maketrans
except AttributeError:
    # fallback for Python 2
    from string import maketrans


def convert_str(data):
    if isinstance(data, str):
        return data
    else:
        data_type = type(data)
        if six.PY3 and data_type == six.binary_type:
            return data.decode('utf-8')
        if data_type in (six.string_types, six.text_type):
            return str(data)
        if data_type == dict:
            data = six.iteritems(data)
            return data_type(map(convert_str, data))
        if data_type in (tuple, list, set):
            return data_type(map(convert_str, data))


def clean_intervals(interval_set):
    newres = []
    for idx, val in enumerate(interval_set):
        newres.append(convert_str(list(interval_set)[idx].data))
    return convert_str(newres)


def get_jxn_genes(jxn, gtree):
    chrom1, pos1, str1, chrom2, pos2, str2, repleft, repright = re.split(':', jxn)
    resL = clean_intervals(gtree[six.b(chrom1)].search(int(pos1)))
    resR = clean_intervals(gtree[six.b(chrom2)].search(int(pos2)))

    # From the left
    genesL = set()
    if len(resL) > 0:
        for idx, val in enumerate(resL):
            Lsymbol = resL[idx]['name2']
            genesL.add(Lsymbol)
    else:
        genesL.add("NA")
    # From the right
    genesR = set()
    if len(resR) > 0:
        for idx, val in enumerate(resR):
            Rsymbol = resR[idx]['name2']
            genesR.add(Rsymbol)
    else:
        genesR.add("NA")
    return (list(genesL), list(genesR))


def overlap_exon(interval, pos, side):
    '''Input is a single interval... list(resL)[0]'''
    try:
        ends = list(map(int, filter(None, interval['exonEnds'].split(","))))
        starts = list(map(int, filter(None, interval['exonStarts'].split(","))))
        frames = list(map(int, filter(None, interval['exonFrames'].split(","))))
        for idx, x in enumerate(starts):
            if pos in range(starts[idx], ends[idx] + 1):  # end is not inclusive so add 1
                if interval['strand'] == "+":
                    if side == 1:
                        dist = ends[idx] - pos
                    else:
                        dist = pos - starts[idx]
                    exon = idx + 1
                else:
                    if side == 1:
                        dist = pos - starts[idx]
                    else:
                        dist = ends[idx] - pos
                    srev = starts[::-1]  # exons are in reverse on "-" strand
                    exon = srev.index(starts[idx]) + 1
                return (dist, exon, frames[idx])
        return("NA", "NA", "NA")
    except:
        return("NA", "NA", "NA")


def get_exons(interval, coding=True, brkpt=False, brkpt_side=1):
    chrom = interval['chrom']
    starts = list(map(int, filter(None, interval['exonStarts'].split(","))))
    ends = list(map(int, filter(None, interval['exonEnds'].split(","))))
    strand = interval['strand']
    trx = interval['name']
    assert len(ends) == len(starts)
    if coding:
        cs = int(interval['cdsStart'])
        ce = int(interval['cdsEnd'])
    else:
        cs = starts[0]
        ce = ends[-1]
    if brkpt:
        if strand == '+':
            if brkpt_side == 1:
                ce = brkpt
            else:
                cs = brkpt
        else:
            if brkpt_side == 1:
                cs = brkpt
            else:
                ce = brkpt
    if strand == '+':
        pos_list = list(zip(range(1, len(starts) + 1), starts, ends))
        found_exons = [(i, s, e) for i, s, e in pos_list if float(min(e, ce) - max(s, cs)) / (e - s) > 0]
    else:
        pos_list = list(zip(range(len(starts), 0, -1), starts, ends))
        found_exons = [(i, s, e) for i, s, e in pos_list if float(min(e, ce) - max(s, cs)) / (e - s) > 0]
    trimmed_exons = [(i, chrom, max(s, cs), min(e, ce), strand, trx) for i, s, e in found_exons]
    # returns list of order, chrom, start, stop, strand, transcript
    return trimmed_exons


def get_jxnside_anno(jxn, gtree, side):
    chrom1, pos1, str1, chrom2, pos2, str2, repleft, repright = re.split(':', jxn)
    if side == 2:
        chrom1 = chrom2
        pos1 = pos2
        flipstr = maketrans("-+", "+-")
        str1 = str2.translate(flipstr)
        repleft = repright  # keep to wiggle later
    if str1 == "+" and side == 1:
        pos1 = int(pos1) - 1
    elif str1 == "-" and side == 1:
        pos1 = int(pos1)
    elif str1 == "+" and side == 2:
        pos1 = int(pos1) - 1
    elif str1 == "-" and side == 2:
        pos1 = int(pos1)

    resL = clean_intervals(gtree[six.b(chrom1)].search(int(pos1)))
    ann = defaultdict(list)

    if len(resL) > 0:
        for idx, val in enumerate(resL):
            ann['symbol'].append(resL[idx]['name2'])  # symbol
            ann['transcript'].append(resL[idx]['name'])
            ann['strand'].append(resL[idx]['strand'])
            dist, exon, frame = overlap_exon(resL[idx], pos1, side)
            ann['exon'].append(exon)
            ann['dist'].append(dist)
            ann['frame'].append(frame)
            ann['cdslen'].append(int(resL[idx]['cdsEnd']) - int(resL[idx]['cdsStart']))

    else:
        ann['symbol'].append("NA")
        ann['transcript'].append("NA")
        ann['strand'].append("NA")
        ann['exon'].append("NA")
        ann['dist'].append("NA")
        ann['frame'].append("NA")
        ann['cdslen'].append("NA")

    # sort results by dist to exon boundary and then by cdslen and then by transcript for consistency
    dsdf = pd.DataFrame(ann).sort_values(['dist', 'cdslen', 'transcript'], ascending=[True, False, True]).reset_index()
    dsdf['dist'] = dsdf['dist'].astype(str)
    ds = dsdf.to_dict(orient='list')

    ann_string = ','.join([str(a) + ":" + b + ":" + c + ":" +
                           str(d) + ":" + str(e) + ":" + str(f) + ':' +
                           str(g) for a, b, c, d, e, f, g in zip(ds['symbol'], ds['transcript'], ds['strand'], ds['exon'], ds['dist'], ds['frame'], ds['cdslen'])])
    return [ds['symbol'][0], ann_string, ds['strand'][0], ds['cdslen'][0] ]  # just the first values for some fields


def get_pos_genes(chrom1, pos1, gtree):
    resL = clean_intervals(gtree[six.b(chrom1)].search(int(pos1)))
    genesL = set()
    if len(resL) > 0:
        for idx, val in enumerate(resL):
            Lsymbol = resL[idx]['name2']
            genesL.add(Lsymbol)
    else:
        genesL.add("NA")
    return list(genesL)


def get_all_exons(jxn, gtree, side, exon_type="trx"):
    chrom1, pos1, str1, chrom2, pos2, str2, repleft, repright = re.split(':', jxn)
    if side == 2:
        chrom1 = chrom2
        pos1 = pos2
        flipstr = maketrans("-+", "+-")
        str1 = str2.translate(flipstr)
        repleft = repright  # keep to wiggle later
    if str1 == "+" and side == 1:
        pos1 = int(pos1) - 1
    elif str1 == "-" and side == 1:
        pos1 = int(pos1)
    elif str1 == "+" and side == 2:
        pos1 = int(pos1) - 1
    elif str1 == "-" and side == 2:
        pos1 = int(pos1)
    resL = clean_intervals(gtree[six.b(chrom1)].search(int(pos1)))
    # From the left

    ann = defaultdict(list)
    if exon_type == "trx":
        if len(resL) > 0:
            for idx, val in enumerate(resL):
                ann['all_exons'].append(get_exons(resL[idx], coding=False))
        else:
            ann['all_exons'].append("NA")
        return ann['all_exons']

    elif exon_type == "fusion":
        if len(resL) > 0:
            for idx, val in enumerate(resL):
               ann['all_exons'].append(get_exons(resL[idx], coding=False, brkpt=pos1, brkpt_side=side))
        else:
            ann['all_exons'].append("NA")
        return ann['all_exons']
    else:
        raise ValueError("Invalid exon_type. Expected one of [trx, fusion]")



#!/usr/bin/env python
# encoding: utf-8

from __future__ import (absolute_import, division, print_function)
import sys
import logging
import subprocess as sp
import re
import starseqr_utils as su

logger = logging.getLogger('STAR-SEQR')


def do_velvet(assemdir, fastq, kmer, errlog, *args):
    ''' Run velvet with single or multiple fastqs '''
    logger.debug('*Running Velvet.')
    velveth_cmd = ['velveth', assemdir, str(kmer), '-short', '-fastq', fastq]
    for ar in args:
        velveth_cmd.extend(['-shortPaired', '-fastq', ar])
    vh_args = list(map(str, velveth_cmd))
    logger.debug('*velveth Command: ' + ' '.join(vh_args))
    try:
        p = sp.Popen(velveth_cmd, stdout=sp.PIPE, stderr=sp.PIPE)
        stdout, stderr = p.communicate()
        if stdout:
            errlog.write(str(stdout))
        if stderr:
            errlog.write(str(stderr))
        if p.returncode != 0:
            logger.error('Error: velveth failed')
    except (OSError) as o:
        logger.error('Exception: ' + str(o))
        logger.error('velveth Failed!', exc_info=True)
        sys.exit(1)
    velvetg_cmd = ['velvetg', assemdir, '-cov_cutoff', '2']
    vg_args = list(map(str, velvetg_cmd))
    logger.debug('*velvetg Command: ' + ' '.join(vg_args))
    try:
        p = sp.Popen(velvetg_cmd, stdout=sp.PIPE, stderr=sp.PIPE)
        stdout, stderr = p.communicate()
        if stdout:
            errlog.write(str(stdout))
        if stderr:
            errlog.write(str(stderr))
        if p.returncode != 0:
            logger.error('Error: velvetg failed')
            sys.exit(1)
    except (OSError) as o:
        logger.error('Exception: ' + str(o))
        logger.error('velvetg Failed!', exc_info=True)
        sys.exit(1)
    # extract header, sequence to generator
    records = su.common.fasta_iter(assemdir + '/contigs.fa')
    return records


def get_assembly_info(jxn, as_type):
    # clean jxn name to write to support folder made previous
    clean_jxn = su.common.safe_jxn(jxn)
    jxn_dir = 'support' + '/' + clean_jxn + '/'

    fusionfq = jxn_dir + 'transcripts_all_fusions.fa'
    fusions_list = list(su.common.fasta_iter(fusionfq))  # list of tuples containing name, seq

    pairfq = jxn_dir + 'span.fastq'
    junctionfq = jxn_dir + 'split.fastq'

    # velvet
    assembly_list = []
    if as_type == 'velvet':
        errlog = open(jxn_dir + 'assembly_log.txt', 'w')
        assembly_list = list(do_velvet(jxn_dir + 'assem_pair', junctionfq, 17, errlog, pairfq))
        errlog.close()

    # confirm assembly crosses breakpoint
    all_crossing = []
    all_seq = []
    all_len = []
    if len(assembly_list) > 0:
        for assembly in assembly_list:
            as_id, as_seq = assembly
            as_len = as_id.split('_')[3]
            # as_cov = int(float(as_id.split('_')[5]))
            all_seq.append(as_seq)
            all_len.append(as_len)
            # print(as_id, as_len, as_cov, as_seq)
            as_crossing_fusions = []
            if len(fusions_list) > 0:
                for fusion in fusions_list:
                    fusion_name, brk = fusion[0].split('|')
                    brk = int(brk)
                    fusion_seq = fusion[1][brk - 10:brk + 10].upper()
                    # print(fusion_name, brk, fusion_seq)
                    # regex solution from: http://stackoverflow.com/questions/2420412/search-for-string-allowing-for-one-mismatch-in-any-location-of-the-string
                    fusion_seq_re = re.compile('|'.join(fusion_seq[:i] + '.{0,2}' +
                                                        fusion_seq[i + 1:] for i in range(len(fusion_seq))))
                    if len(fusion_seq_re.findall(as_seq.upper())) or len(fusion_seq_re.findall(su.common.rc(as_seq).upper())) > 0:
                        as_crossing_fusions.append(fusion_name)
                all_crossing.extend(as_crossing_fusions)
    # convert all lists to string for good recordkeeping
    all_seq = ','.join(list(map(str, all_seq)))
    all_len = ','.join(list(map(str, all_len)))
    all_crossing = ','.join(list(map(str, all_crossing)))
    return all_seq, all_len, all_crossing



#!/usr/bin/env python

import unittest
import ssw


class TestSSW(unittest.TestCase):
    def test_mixed_case(self):
        reference = "GTGCGATGTGCGATGAGATC"
        query = reference.lower()
        aligner = ssw.Aligner()
        al = aligner.align(query, reference)
        self.assertEqual(al.match_count, len(reference))
        self.assertEqual(al.mismatch_count, 0)
        self.assertEqual(al.insertion_count, 0)
        self.assertEqual(al.deletion_count, 0)
        self.assertEqual(al.cigar, '20M')

    def test_perfect_alignment(self):
        reference = "GTGCGATGTGCGATGAGATC"
        query = reference
        aligner = ssw.Aligner()
        al = aligner.align(query, reference)
        self.assertEqual(al.match_count, len(reference))
        self.assertEqual(al.mismatch_count, 0)
        self.assertEqual(al.insertion_count, 0)
        self.assertEqual(al.deletion_count, 0)
        self.assertEqual(al.cigar, '20M')

    def test_rc_alignment(self):
        reference = "GTGCGATGTGCGATGAGATC"
        query = "GATCTCATCGCACATCGCAC"
        aligner = ssw.Aligner()
        al = aligner.align(query, reference)
        self.assertEqual(al.match_count, 20)
        self.assertEqual(al.mismatch_count, 0)
        self.assertEqual(al.insertion_count, 0)
        self.assertEqual(al.deletion_count, 0)
        self.assertEqual(al.cigar, "20M")

    def test_insertion(self):
        reference = "GTGCGATGTGCGATGAGATC"
        query = reference[:10] + 'A' + reference[10:]
        aligner = ssw.Aligner()
        al = aligner.align(query, reference)
        self.assertEqual(al.match_count, 20)
        self.assertEqual(al.mismatch_count, 0)
        self.assertEqual(al.insertion_count, 1)
        self.assertEqual(al.deletion_count, 0)
        self.assertEqual(al.cigar, "10M1I10M")

    def test_mismatch(self):
        reference = "GTGCGATGTGCGATGAGATC"
        query = reference[:9] + 'A' + reference[10:]
        aligner = ssw.Aligner()
        al = aligner.align(query, reference)
        self.assertEqual(al.match_count, 19)
        self.assertEqual(al.mismatch_count, 1)
        self.assertEqual(al.insertion_count, 0)
        self.assertEqual(al.deletion_count, 0)
        self.assertEqual(al.cigar, "20M")

    def test_deletion(self):
        reference = "GTGCGATGTGCGATGAGATC"
        query = reference[:10] + reference[11:]
        aligner = ssw.Aligner()
        al = aligner.align(query, reference)
        self.assertEqual(al.match_count, 19)
        self.assertEqual(al.mismatch_count, 0)
        self.assertEqual(al.insertion_count, 0)
        self.assertEqual(al.deletion_count, 1)
        self.assertEqual(al.cigar, "10M1D9M")

    def test_issue_1(self):
        # https://github.com/vishnubob/ssw/issues/1
        reference = "CCC" + "AGCT" * 10
        query = "AGGT" * 10
        aligner = ssw.Aligner()
        alignment = aligner.align(query, reference)
        (r_line, m_line, q_line) = alignment.alignment
        self.assertEqual(r_line, "AGCTAGCTAGCTAGCTAGCTAGCTAGCTAGCTAGCTAG")
        self.assertEqual(m_line, "||*|||*|||*|||*|||*|||*|||*|||*|||*|||")
        self.assertEqual(q_line, "AGGTAGGTAGGTAGGTAGGTAGGTAGGTAGGTAGGTAG")

    def test_issue_3(self):
        # https://github.com/vishnubob/ssw/issues/3
        self.assertRaises(ValueError, ssw.Aligner, gap_open=1, gap_extend=2)
        self.assertRaises(ValueError, ssw.Aligner, gap_open=1, gap_extend=1)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8

from __future__ import (absolute_import, division, print_function)
import logging
import primer3
import starseqr_utils as su

logger = logging.getLogger("STAR-SEQR")


def runp3(seq_id, sequence, target=None):
    ''' can either use ":" in sequence, a base target or choose the middle of a read'''
    # get index of split to design targets
    if target:
        if (target + 30  > len(sequence)) or (target - 30 < 0 ):
            # target to close to border for primer generation
            return ()
        brk_target = [target - 10, 20]
    else:
        if ":" in sequence:
            mybrk = int(sequence.index(":"))
            sequence = sequence.replace(":", "")
            if (mybrk + 30  > len(sequence)) or (mybrk - 30 < 0 ):
                # target to close to border for primer generation
                return ()
            brk_target = [mybrk - 10, 20]
        else:
            if len(str(sequence)) < 75:
                return ()
            brk_target = [int(len(sequence) / 2), 1]
    # default values
    mydres = {
        'SEQUENCE_ID': seq_id,
        'SEQUENCE_TEMPLATE': sequence,
        # 'SEQUENCE_INCLUDED_REGION': [],
        'SEQUENCE_TARGET': brk_target  # start, len
    }
    mypres = {
        'PRIMER_NUM_RETURN': 1,
        'PRIMER_TASK': 'generic',
        'PRIMER_PICK_LEFT_PRIMER': 1,
        'PRIMER_PICK_RIGHT_PRIMER': 1,
        'PRIMER_PICK_INTERNAL_OLIGO': 1,
        'PRIMER_OPT_SIZE': 20,
        'PRIMER_MIN_SIZE': 18,
        'PRIMER_MAX_SIZE': 27,
        'PRIMER_OPT_TM': 60.0,
        'PRIMER_MIN_TM': 55.0,
        'PRIMER_MAX_TM': 65.0,
        'PRIMER_OPT_GC_PERCENT': 50,
        'PRIMER_MIN_GC': 20.0,
        'PRIMER_MAX_GC': 80.0,
        'PRIMER_LIBERAL_BASE': 0,
        'PRIMER_MIN_THREE_PRIME_DISTANCE': 0,  # can be 0=unique, or -1 multiple of same
        'PRIMER_LOWERCASE_MASKING': 0,  # 0 allows all case, 1, rejects if near 3'.
        'PRIMER_MAX_POLY_X': 100,
        'PRIMER_INTERNAL_MAX_POLY_X': 100,
        'PRIMER_INTERNAL_MAX_SELF_END': 8,
        'PRIMER_SALT_MONOVALENT': 50.0,
        'PRIMER_DNA_CONC': 50.0,
        'PRIMER_MAX_NS_ACCEPTED': 0,
        'PRIMER_MAX_SELF_ANY': 12,
        'PRIMER_MAX_SELF_END': 8,
        'PRIMER_PAIR_MAX_COMPL_ANY': 12,
        'PRIMER_PAIR_MAX_COMPL_END': 8,
        'PRIMER_PAIR_MAX_DIFF_TM': 6,
        'PRIMER_PRODUCT_SIZE_RANGE': [[75, 125], [125, 150],
                                      [150, 200], [200, 300]],
    }
    try:
        p3output = primer3.bindings.designPrimers(mydres, mypres)
        if (p3output['PRIMER_PAIR_NUM_RETURNED'] < 1):
            return ()
        else:
            p3res = parsep3(p3output)
    except (OSError):
        logger.error("Primer Design Failed- continuing", exc_info=True)
        return ()
    except (Exception):
        return ()
    return p3res


def parsep3(p3output):
    # logger.debug('Parsing Primer3 Results')
    Lprimer = str(p3output['PRIMER_LEFT_0_SEQUENCE'])
    # Ltm = round(p3output['PRIMER_LEFT_0_TM'])
    # Ltuple = p3output['PRIMER_LEFT_0']
    # Lstart, Llen = str(Ltuple[0]), str(Ltuple[1])
    Rprimer = str(p3output['PRIMER_RIGHT_0_SEQUENCE'])
    # Rtm = round(p3output['PRIMER_RIGHT_0_TM'])
    # Rtuple = p3output['PRIMER_RIGHT_0']
    # Rstart, Rlen = str(Rtuple[0]), str(Rtuple[1])
    # amplen = str(int(Rstart) - int(Lstart))
    return (Lprimer.upper(), Rprimer.upper())


def wrap_runp3(jxn, cross_fusions):
    '''
    use the breakpoint info stored in the fasta file header.
    Design primers using known cross fusion transcripts
    '''
    clean_jxn = su.common.safe_jxn(jxn)
    jxn_dir = 'support' + '/' + clean_jxn + '/'

    fusionfq = jxn_dir + 'transcripts_all_fusions.fa'
    fusions_list = list(su.common.fasta_iter(fusionfq))  # list of tuples containing name, seq
    cross_fusions = [x for x in cross_fusions.split(',') if x] # removes empty strings
    if len(fusions_list) > 0 and len(cross_fusions) > 0:
        for fusion in fusions_list:
            fusion_name, brk = fusion[0].split('|')
            fus_seq = fusion[1]
            brk = int(brk)
            if cross_fusions[0] == fusion_name:
                return runp3(fusion_name, fus_seq.upper(), brk)
    else:
        return ()



#!/usr/bin/env python
# encoding: utf-8

from __future__ import (absolute_import, division, print_function)
import os
import subprocess
import logging
import gzip
from intervaltree_bio import GenomeIntervalTree, UCSCTable


logger = logging.getLogger("STAR-SEQR")


# modified from https://github.com/chapmanb/cloudbiolinux/blob/master/utils/prepare_tx_gff.py
def file_exists(fname):
    """Check if a file exists and is non-empty."""
    try:
        return fname and os.path.exists(fname) and os.path.getsize(fname) > 0
    except OSError:
        return False


def gtf_to_genepred(gtf, out_file):
    if file_exists(out_file):
        logger.warning("Skipping gtf_to_genepred as files already exist!")
        return out_file
    cmd = "gtfToGenePred -allErrors -ignoreGroupsWithoutExons -genePredExt -geneNameAsName2 {gtf} {out_file}"
    subprocess.check_call(cmd.format(**locals()), shell=True)
    return out_file


def genepred_to_UCSCtable(genepred, out_file):
    # header = ["#bin", "name", "chrom", "strand",
    #           "txStart", "txEnd", "cdsStart", "cdsEnd",
    #           "exonCount", "exonStarts", "exonEnds", "score",
    #           "name2", "cdsStartStat", "cdsEndStat",
    #           "exonFrames"]
    if file_exists(out_file):
        logger.warning("Skipping genepred_to_UCSC_table as files already exist!")
        return out_file
    with open(genepred) as in_handle, gzip.open(out_file, "wt") as out_handle:
        counter = -1
        current_item = None
        for l in in_handle:
            item = l.split("\t")[0]
            if current_item != item:
                current_item = item
                counter = counter + 1
            out_handle.write("\t".join([str(counter), l]))
    return out_file


def gtf2tree(gtf_path):
    genepred_annot = os.path.splitext(gtf_path)[0] + ".genePred"
    ucsc_annot = os.path.splitext(gtf_path)[0] + ".UCSCTable.gz"
    gtf_to_genepred(gtf_path, genepred_annot)
    genepred_to_UCSCtable(genepred_annot, ucsc_annot)
    kg = gzip.open(ucsc_annot)
    gtree = GenomeIntervalTree.from_table(fileobj=kg, mode='tx', parser=UCSCTable.ENS_GENE)
    return gtree



#!/usr/bin/env python
# encoding: utf-8

from __future__ import (absolute_import, division, print_function)
from six import reraise as raise_
import sys
import time
import pandas as pd
import logging
import svtools_star.svtools.bedpetovcf as sv  # temporary until svtools is fixed

try:
    maketrans = ''.maketrans
except AttributeError:
    # fallback for Python 2
    from string import maketrans

logger = logging.getLogger("STAR-SEQR")


def write_header(args, fh, file_type):
    '''Follow the spec here: https://samtools.github.io/hts-specs/VCFv4.2.pdf'''
    today = time.strftime('%m/%d/%Y')
    vpcmd = str(' '.join(sys.argv))
    vcf_head = '\t'.join(['#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT', args.prefix])
    bedpe_head = '\t'.join(['#CHROM_A', 'START_A', 'END_A', 'CHROM_B', 'START_B', 'END_B', 'ID', 'QUAL', 'STRAND_A', 'STRAND_B',
                            'TYPE', 'FILTER', 'NAME_A', 'REF_A', 'ALT_A', 'NAME_B', 'REF_B', 'ALT_B', 'INFO_A', 'INFO_B', 'FORMAT', args.prefix])
    header = ['##fileformat=VCFv4.2', '##fileDate=' + today,
              '##source=STAR-SEQR', '##reference=' + args.fasta,
              '##FILTER=<ID=PASS,Description="All filters passed">',
              '##FILTER=<ID=FAIL,Description="Site failed to reach confidence">',
              '##INFO=<ID=ANN,Number=.,Type=String,Description="Functional annotations: \'Symbol\'">',
              '##INFO=<ID=IMPRECISE,Number=0,Type=Flag,Description="Imprecise structural variation">',
              '##INFO=<ID=CIPOS,Number=2,Type=Integer,Description="Confidence interval around POS for imprecise variants">',
              '##INFO=<ID=CIEND,Number=2,Type=Integer,Description="Confidence interval around END for imprecise variants">',
              '##INFO=<ID=END,Number=1,Type=Integer,Description="End position of the variant described in this record">',
              '##INFO=<ID=HOMLEN,Number=-1,Type=Integer,Description="Length of base pair identical micro-homology at event breakpoints">',
              '##INFO=<ID=HOMSEQ,Number=.,Type=String,Description="Sequence of base pair identical micro-homology at event breakpoints">',
              '##INFO=<ID=MATEID,Number=.,Type=String,Description="ID of mate breakends">',
              '##INFO=<ID=EVENT,Number=1,Type=String,Description="ID of event associated to breakend">',
              '##INFO=<ID=SECONDARY,Number=0,Type=Flag,Description="Secondary breakend in a multi-line variants">',
              '##INFO=<ID=SEQLEN,Number=.,Type=String,Description="Length of assembled sequence of reads supporting breakpoint">',
              '##INFO=<ID=SVLEN,Number=-1,Type=Integer,Description="Difference in length between REF and ALT alleles">',
              '##INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of structural variant">',
              '##ALT=<ID=DEL,Description="Deletion">',
              '##ALT=<ID=DUP,Description="Duplication">',
              '##ALT=<ID=DUP:TANDEM,Description="Tandem Duplication">',
              '##ALT=<ID=INS,Description="Insertion of novel sequence">',
              '##ALT=<ID=INV,Description="Inversion">',
              '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">',
              '##FORMAT=<ID=SP,Number=1,Type=String,Description="Number of paired discordant spanning reads">',
              '##FORMAT=<ID=JXL,Number=1,Type=String,Description="Number of reads coming from the left">',
              '##FORMAT=<ID=JXR,Number=1,Type=String,Description="Number of reads coming from the right">',
              '##FORMAT=<ID=GQ,Number=1,Type=String,Description="Genotype Quality">',
              '##variants_justified=left',
              '##STAR-SV_CMD=' + vpcmd]
    if file_type == "bedpe":
        header.append(bedpe_head)
    else:
        header.append(vcf_head)
    header_out = '\n'.join(header)
    print(header_out, file=fh)


def get_svtype_func(str1, str2, c1, c2):
    # Translocations are represented as BNDs
    if c1 != c2:
        svtype = "BND"
    else:
        # STAR notation is same as other tools after strand2 is flipped.
        if str(str1) == "+" and str(str2) == "+":
            svtype = "INV"
        elif str(str1) == "-" and str(str2) == "-":
            svtype = "INV"
        elif str(str1) == "+" and str(str2) == "-":
            svtype = "DEL"
        elif str(str1) == "-" and str(str2) == "+":
            svtype = "DUP"
    return svtype


def get_bnd_alt_func(c1, c2, s1, s2, str1, str2, which_bnd):
    if which_bnd == 'A':
        if str(str1) == "+" and str(str2) == "+":
            alt = "N" + "]" + str(c2) + ":" + str(s2) + "]"
        elif str(str1) == "-" and str(str2) == "-":
            alt = "[" + str(c2) + ":" + str(s2) + "[" + "N"
        elif str(str1) == "+" and str(str2) == "-":
            alt = "N" + "[" + str(c2) + ":" + str(s2) + "["
        elif str(str1) == "-" and str(str2) == "+":
            alt = "]" + str(c2) + ":" + str(s2) + "]" + "N"
    elif which_bnd == 'B':
        if str(str1) == "+" and str(str2) == "+":
            alt = "N" + "]" + str(c1) + ":" + str(s1) + "]"
        elif str(str1) == "-" and str(str2) == "-":
            alt = "[" + str(c1) + ":" + str(s1) + "[" + "N"
        elif str(str1) == "+" and str(str2) == "-":
            alt = "]" + str(c1) + ":" + str(s1) + "]" + "N"
        elif str(str1) == "-" and str(str2) == "+":
            alt = "N" + "[" + str(c1) + ":" + str(s1) + "["
    return alt


def get_bedpestuff_func(df_in, svtype):
    if svtype == "BND":
        df = df_in[df_in['svtype'] == "BND"]
        if len(df.index) > 0:
            df['REF_A'] = "N"
            df['ALT_A'] = df.apply(lambda x: get_bnd_alt_func(x['c1'], x['c2'], x['s1'], x['s2'], x['st1'], x['st2'], "A"), axis=1)
            df['NAME_A'] = df['id'].astype(int).astype(str) + "_1"
            df['NAME_B'] = df['id'].astype(int).astype(str) + "_2"
            df['REF_B'] = "N"
            df['ALT_B'] = df.apply(lambda x: get_bnd_alt_func(x['c1'], x['c2'], x['s1'], x['s2'], x['st1'], x['st2'], "B"), axis=1)
            df['INFO_A'] = "SVTYPE=" + df['svtype'] + ';' + \
                "POS=" + df['s1'].astype(str) + ';' + \
                "END=" + df['s2'].astype(str) + ';' + \
                "ANN=" + df['ann'] + ';' + \
                "CIPOS=" + df['a2'] + ',' + df['a1'] + ';' + \
                "HOMLEN=" + df['a1'] + ';' + \
                "MATEID=" + df['NAME_B'] + ';' + \
                "EVENT=" + df['id'].astype(str)
                # "SEQLEN=" + df['velvet'].str.len().astype(str)
            df['INFO_B'] = "SVTYPE=" + df['svtype'] + ';' + \
                "POS=" + df['s1'].astype(str) + ';' + \
                "END=" + df['s2'].astype(str) + ';' + \
                "ANN=" + df['ann'] + ';' + \
                "CIPOS=" + df['a1'] + ',' + df['a2'] + ';' + \
                "HOMLEN=" + df['a2'] + ';' + \
                "MATEID=" + df['NAME_A'] + ';' + \
                "EVENT=" + df['id'].astype(str) + ';' + \
                "SECONDARY"
                # "SEQLEN=" + df['velvet'].str.len().astype(str)
    else:
        df = df_in[df_in['svtype'] != "BND"]
        if len(df.index) > 0:
            df['REF_A'] = "N"
            df['ALT_A'] = "<" + df['svtype'] + ">"
            df['NAME_A'] = df['id'].astype(int).astype(str)
            df['NAME_B'] = "."
            df['REF_B'] = "."
            df['ALT_B'] = "."
            df['INFO_A'] = "SVTYPE=" + df['svtype'] + ';' + \
                "POS=" + df['s1'].astype(str) + ';' + \
                "END=" + df['s2'].astype(str) + ';' + \
                "ANN=" + df['ann'] + ';' + \
                "CIPOS=" + df['a1'] + "," + df['a2'] + ';' + \
                "HOMLEN=" + df['a2'] + ';' + \
                "SVLEN=" + df['dist'].astype(str)
            df['INFO_B'] = "."
    return df


def write_bedpe(df, out_bedpe, args):
    '''the bedpe should be 0 based, incoming values are 0 based.'''
    bedpe_fh = open(out_bedpe, 'w')
    write_header(args, bedpe_fh, "bedpe")
    if len(df.index) >= 1:
        df = df.reset_index(drop=True)
        df = df.reset_index()
        df['c1'], df['s1'], df['st1'], df['c2'], df['s2'], df['st2'], df['a1'], df['a2'] = zip(*df['name'].str.split(':').tolist())
        # STAR has second object flipped for DNA
        flipstr = maketrans("-+", "+-")
        df['st2'] = df['st2'].str.translate(flipstr)
        # use 0-based positions from breakpoint cols
        df['c1'], df['s1'], df['st1'] = zip(*df['breakpoint_left'].str.split(':').tolist())
        df['c2'], df['s2'], df['st2'] = zip(*df['breakpoint_right'].str.split(':').tolist())
        df['s1'] = df['s1'].astype(int)
        df['s2'] = df['s2'].astype(int)
        # Common stuff to BNDs and others
        df['e1'] = df['s1'] + 1  # these are 1-based endings.
        df['e2'] = df['s2'] + 1
        df['id'] = df['index'].astype(int) + 1
        df['qual'] = df['jxn_first'] + df['jxn_second']  # Eventually get a handle on a scoring metric besides number of reads
        df['svtype'] = df.apply(lambda x: get_svtype_func(x['st1'], x['st2'], x['c1'], x['c2']), axis=1)
        df['filter'] = "PASS"

        # Get unique to BNDs and others then merge
        pd.options.mode.chained_assignment = None  # default='warn'
        df_bnd = get_bedpestuff_func(df, "BND")
        df_nonbnd = get_bedpestuff_func(df, "nonBND")
        df2 = pd.concat([df_bnd, df_nonbnd], ignore_index=True).sort_values(['c1', 's1'], ascending=[True, True])   #

        df2['FORMAT'] = "GT:SP:JXL:JXR"
        df2[args.prefix] = "./." + ":" + df2['spans_disc'].astype(str) + ":" + df2['jxn_first'].astype(str) + ":" + df2['jxn_second'].astype(str)
        outcols = ['c1', 's1', 'e1', 'c2', 's2', 'e2', 'id', 'qual', 'st1', 'st2', 'svtype', 'filter',
                   'NAME_A', 'REF_A', 'ALT_A', 'NAME_B', 'REF_B', 'ALT_B', 'INFO_A', 'INFO_B', 'FORMAT', args.prefix]
        df2.to_csv(path_or_buf=bedpe_fh, header=False, sep='\t', columns=outcols, mode='a', index=False)
    logger.info("bedpe creation was successful!")


def write_vcf(in_bed, out_vcf):
    try:
        in_fh = open(in_bed, 'r')
        out_fh = open(out_vcf, 'w')
        sv.bedpeToVcf(in_fh, out_fh)
    except (OSError) as o:
        logger.error("Exception: " + str(o))
        logger.error("sv Failed", exc_info=True)
        traceback = sys.exc_info()[2]
        raise_(ValueError, e, traceback)
    logger.info("VCF creation was successful!")


def process(bkpt_df, args):
    bedpe_path = args.prefix + '_STAR-SEQR_breakpoints.bedpe'
    vcf_path = args.prefix + '_STAR-SEQR_breakpoints.vcf'
    write_bedpe(bkpt_df, bedpe_path, args)
    write_vcf(bedpe_path, vcf_path)



import sys

class InputStream(object):
    '''This class handles opening either stdin or a gzipped or non-gzipped file'''

    def __init__(self, string):
        '''Create a new wrapper around a stream'''
        if string in (None, '-', 'stdin') and self.valid(string):
            self.handle = sys.stdin
        elif string.endswith('.gz'):
            import gzip
            self.handle = gzip.open(string, 'rb')
        else:
            self.handle = open(string, 'r')

    @staticmethod
    def valid(string):
        '''Check if we allow the opening of stdin'''
        if string is None and sys.stdin.isatty():
            raise IOError('no input specified but terminal is interactive')
        return True
    
    def __enter__(self):
        '''Support use of with by passing back the originating handle'''
        return self.handle

    def __exit__(self, *kwargs):
        '''Support use of with by closing on exit of the context'''
        self.handle.close()

    def __iter__(self):
        '''Support use in loops like a normal file object'''
        return self.handle.__iter__()

    def close(self):
        '''Close the underlying handle'''
        return self.handle.close()



#!/usr/bin/env python
# encoding: utf-8

from __future__ import (absolute_import, division, print_function)
import os
import sys
import time
import argparse
import pandas as pd
import numpy as np
import logging
from collections import OrderedDict
import starseqr_utils as su


def parse_args():
    class FullPaths(argparse.Action):
        """Expand user- and relative-paths"""
        def __call__(self, parser, namespace, values, option_string=None):
            setattr(namespace, self.dest, os.path.abspath(os.path.expanduser(values)))

    usage = " "
    parser = argparse.ArgumentParser(description="STAR-SEQR Parameters:", epilog=usage)
    # create STAR alignment
    group1 = parser.add_argument_group('Do Alignment', '')
    group1.add_argument('-1', '--fastq1', type=str, required=False, action = FullPaths,
                        help='fastq.gz 1')
    group1.add_argument('-2', '--fastq2', type=str, required=False, action = FullPaths,
                        help='fastq.gz 2')
    group1.add_argument('-i', '--star_index', type=str, required=False, action=FullPaths,
                        help='path to STAR index folder')
    group1.add_argument('-m', '--mode', type=int, required=False,
                        default=0,
                        choices=[0, 1],
                        help='STAR alignment sensitivity Mode. 0=Default, 1=More-Sensitive')

    # existing STAR alignment
    group2 = parser.add_argument_group('Use Existing Alignment', '')
    group2.add_argument('-sj', '--star_jxns', type=str, required=False, action = FullPaths,
                        help='chimeric junctions file produce by STAR')
    group2.add_argument('-ss', '--star_sam', type=str, required=False, action = FullPaths,
                        help='chimeric sam file produced by STAR')

    # DNA Parameters
    group3 = parser.add_argument_group('DNA Parameters', '')
    group3.add_argument('-d', '--dist', type=int, required=False,
                        default=100000,
                        help='DNA Only: minimum distance to call junctions')
    group3.add_argument('-j', '--jxn_reads', type=int, required=False,
                        default=2,
                        help='DNA Only: minimum number of junction reads to keep junctions.')
    group3.add_argument('-s', '--span_reads', type=int, required=False,
                        default=2,
                        help='DNA Only: minimum number of spanning discordant read pairs to call junctions.')

    # shared args
    parser.add_argument('-p', '--prefix', type=str, required=True,
                        help='prefix to name files')
    parser.add_argument('-r', '--fasta', type=str, required=True, action = FullPaths,
                        help='indexed fasta (.fa|.fa.gz)')
    parser.add_argument('-g', '--gtf', type=str, required=True, action = FullPaths,
                        help='gtf file. (.gtf|.gtf.gz)')
    parser.add_argument('-n', '--nucleic_type', type=str, required=False,
                        default="RNA",
                        help='nucleic acid type',
                        choices=["RNA", "DNA"])
    parser.add_argument('-t', '--threads', type=int, required=False,
                        default=8,
                        help='Number of threads to use for STAR and STAR-SEQR. 4-8 recommended.')
    parser.add_argument('-b', '--bed_file', type=str, required=False, action = FullPaths,
                        help='Bed file to subset analysis')
    parser.add_argument('--subset_type', type=str, required=False,
                        default="either",
                        help='allow fusions to pass with either one breakend in bed file or require both. Must use -b.',
                        choices=["either", "both"])
    parser.add_argument('-a', '--as_type', type=str, required=False,
                        default="velvet",
                        help='nucleic acid type',
                        choices=["velvet"])
    parser.add_argument('--keep_dups', action='store_true',
                        help='keep read duplicates. Use for PCR data or with molecular tags')
    parser.add_argument('--keep_gene_dups', action='store_true',
                        help='allow internal gene duplications to be considered')
    parser.add_argument('--keep_mito', action='store_true',
                        help='allow RNA fusions to contain at least one breakpoint from Mitochondria')
    parser.add_argument('-v', '--verbose', action="count",
                        help="verbose level... repeat up to three times.")
    args = parser.parse_args()

    # check that the correct args have been specified
    align = [args.fastq1, args.fastq2, args.star_index]
    call = [args.star_jxns, args.star_sam]
    if any(align) and any(call):
        print("Error: Please choose either fastqs or STAR existing files as input!")
        sys.exit(1)
    if any(align) and None in align:
        print("Error: Fastq1, Fastq2, and the STAR index must be specified if doing alignment")
        sys.exit(1)
    if any(call) and None in call:
        print("Error: The STAR .junctions and .sam file must be specified if using existing alignment")
        sys.exit(1)
    return args


def set_log_level_from_verbose(ch, args):
    if not args.verbose:
        ch.setLevel('ERROR')
    elif args.verbose == 1:
        ch.setLevel('WARNING')
    elif args.verbose == 2:
        ch.setLevel('INFO')
    elif args.verbose >= 3:
        ch.setLevel('DEBUG')


def apply_choose_order(df):
    df['order'] = df.apply(lambda x: su.core.choose_order(x['chrom1'], x['pos1'], x['chrom2'], x['pos2']), axis=1)
    return df


def apply_normalize_jxns(df):
    df['name'] = df.apply(lambda x: su.core.normalize_jxns(x['chrom1'], x['chrom2'], x['pos1'], x['pos2'],
                                                           x['str1'], x['str2'], x['jxnleft'], x['jxnright'],
                                                           x['order']), axis=1)
    return df


def apply_count_vals(df):
    new_df = df.groupby(['name', 'order'], as_index=True).agg({'readid': {'reads': lambda col: ','.join(
        col), 'counts': 'count'}}).reset_index().pivot(index='name', columns='order').reset_index()
    return new_df


def apply_pairs_func(args):
    df, dd = args
    df['spans'], df['spanreads'] = zip(*df.apply(lambda x: su.core.get_pairs_func(x['name'], dd), axis=1))
    return df


def apply_jxn_strand(df):
    _, _, df['test_strand'], _ = zip(*df.apply(lambda x: su.annotate_sv.get_jxnside_anno(x['name'], gtree, 1), axis=1))
    return df


def apply_flip_func(df):
    df['name'], df['flip'] = zip(*df.apply(lambda x: su.core.flip_jxn(x['name'], x['test_strand']), axis=1))
    return df


def wrap_annotate_junctions(df):
    df['ann_format'] = "Symbol:Transcript:Strand:Exon_No:Dist_to_Exon:Frame:CDS_Length"
    left_ann = df.apply(lambda x: su.annotate_sv.get_jxnside_anno(x['name'], gtree, 1), axis=1)
    df['left_symbol'], df['left_annot'], df['left_strand'], df['left_cdslen'] = zip(*left_ann)
    right_ann = df.apply(lambda x: su.annotate_sv.get_jxnside_anno(x['name'], gtree, 2), axis=1)
    df['right_symbol'], df['right_annot'], df['right_strand'], df['right_cdslen'] = zip(*right_ann)
    df['ann'] = df['left_symbol'] + "--" + df['right_symbol']
    return df


def wrap_jxn_info(df):
    df['left_canonical'] = df['left_annot'].str.split(':', expand=True)[4].apply(lambda x: 'CANONICAL_SPLICING' if x == '0' else 'NON-CANONICAL_SPLICING')
    df['right_canonical'] = df['right_annot'].str.split(':', expand=True)[4].apply(lambda x: 'CANONICAL_SPLICING' if x == '0' else 'NON-CANONICAL_SPLICING')
    df['splice_type'] = np.where((df['left_canonical'] == 'CANONICAL_SPLICING') &
                                 (df['right_canonical'] == 'CANONICAL_SPLICING'), 'CANONICAL_SPLICING', 'NON-CANONICAL_SPLICING')
    # get all genes associated to look for overlap for each read later..
    df['left_all'], df['right_all'] = zip(*df.apply(lambda x: su.annotate_sv.get_jxn_genes(x['name'], gtree), axis=1))
    df['txunion'] = [list(set(a).union(set(b))) for a, b in zip(df.left_all, df.right_all)]
    df['txintersection'] = [list(set(a).intersection(set(b))) for a, b in zip(df.left_all, df.right_all)]
    # Get reference dist
    df['dist'] = df.apply(lambda x: su.core.get_distance(x['name']), axis=1)
    return df


def apply_get_rna_support(args):
    df, in_bam = args
    dict_res = list(df.apply(lambda x: su.support_funcs_rna.get_rna_support(x['name'], x['txunion'], x['supporting_reads'], in_bam, gtree), axis=1))
    newdf = pd.DataFrame.from_records(dict_res, index='name')
    return newdf  # not passed back in same df


def wrap_exons2seq(args):
    df, fa_object = args
    df['left_trx_exons'] = df.apply(lambda x: su.annotate_sv.get_all_exons(x['name'], gtree, 1, exon_type="trx"), axis=1)
    df['right_trx_exons'] = df.apply(lambda x: su.annotate_sv.get_all_exons(x['name'], gtree, 2, exon_type="trx"), axis=1)
    df['left_fusion_exons'] = df.apply(lambda x: su.annotate_sv.get_all_exons(x['name'], gtree, 1, exon_type="fusion"), axis=1)
    df['right_fusion_exons'] = df.apply(lambda x: su.annotate_sv.get_all_exons(x['name'], gtree, 2, exon_type="fusion"), axis=1)
    df.apply(lambda x: su.core.exons2seq(fa_object, x['left_trx_exons'], x['name'], "left"), axis=1)
    df.apply(lambda x: su.core.exons2seq(fa_object, x['right_trx_exons'], x['name'], "right"), axis=1)
    df.apply(lambda x: su.core.exons2seq(fa_object, x['left_fusion_exons'], x['name'], "all_fusions", x['right_fusion_exons']), axis=1)
    df['write_seq'] = "Finished"
    return df['write_seq'] # sequences are written to fasta not passed, but need to pass something


def apply_primers_func(df):
    df['primers'] = df.apply(lambda x: su.run_primer3.wrap_runp3(x['name'], x['assembly_cross_fusions']), axis=1).apply(lambda x: ",".join(x))
    return df


def apply_get_cross_homology(df):
    df['span_homology_score'], df['jxn_homology_score'] = zip(*df.apply(lambda x: su.cross_homology.get_cross_homology(x['name']), axis=1))
    return df


def apply_get_diversity(df):
    df['overhang_diversity'] = df.apply(lambda x: su.overhang_diversity.get_diversity(x['name']), axis=1)
    return df


def apply_get_minfrag_length(df):
    df['minfrag20'], df['minfrag35'] = zip(*df.apply(lambda x: su.core.get_minfrag_length(x['name'], x), axis=1))
    return df


def apply_get_assembly_info(args):
    df, as_type = args
    df['assembly'], df['assembly_len'], df['assembly_cross_fusions'] = zip(
        *df.apply(lambda x: su.run_assembly.get_assembly_info(x['name'], as_type), axis=1))
    return df


def apply_get_fusion_class(df):
    df['Fusion_Class'] = df.apply(lambda x: su.core.get_fusion_class(x['name'], x['txintersection']), axis=1)
    return df


def apply_get_annot_db(args):
    df, dbargs = args
    chimerdb3, fuca = dbargs
    df['ChimerDB_ann'] = df.apply(lambda x: su.annotate_db.get_chimerdb(x['name'], chimerdb3), axis=1)
    df['FusionCancer_ann'] = df.apply(lambda x: su.annotate_db.get_fusioncancerdb(x['name'], fuca), axis=1)
    return df


def rna_closeout(prefix, stats_res, breakpoints_fh):
    # Write stats to file
    stats_fh = open(prefix + "_STAR-SEQR.stats", 'w')
    for key, value in stats_res.items():
        stats_fh.write(key + "\t" + str(value) + "\n")
    breakpoints_fh.close()
    brkpt_path = prefix + "_STAR-SEQR_breakpoints.txt"
    bedpe_path = prefix + "_STAR-SEQR_breakpoints.bedpe"
    su.core.write_bedpe(brkpt_path, bedpe_path)


def main():
    start = time.time()
    args = parse_args()

    # file log
    cust_format = '%(asctime)s - %(module)s - %(levelname)s - %(message)s'
    logging.basicConfig(level=logging.DEBUG, format=cust_format,
                        datefmt="%Y-%m-%d %H:%M:%S", filename=args.prefix + '_STAR-SEQR.log')
    # console log
    logger = logging.getLogger("STAR-SEQR")
    console = logging.StreamHandler()
    set_log_level_from_verbose(console, args)
    formatter = logging.Formatter('%(asctime)-15s - %(levelname)-4s - %(message)s', "%Y-%m-%d %H:%M")
    console.setFormatter(formatter)
    logger.addHandler(console)

    # start analysis
    logger.info("***************STAR-SEQR******************")
    logger.info("CMD = " + str(' '.join(sys.argv)))
    logger.info("STAR-SEQR_version = " + str(su.__version__))
    logger.info('Starting to work on sample: ' + args.prefix)

    # check dependent software can be found
    depend_tools = ['samtools', 'bamfilternames', 'velveth', 'gtfToGenePred']
    for tool in depend_tools:
        if not su.common.which(tool):
            logger.error(tool + "exe not found on path! Quitting.")
            sys.exit(1)

    # check files exist and get abs paths. Necessary since we switch folders.
    depend_paths = [args.fasta, args.gtf, args.fastq1, args.fastq2,
                    args.bed_file, args.star_jxns, args.star_sam]
    for f_item in depend_paths:
        if f_item:
            su.common.check_file_exists(f_item)
            logger.info("Found input: %s", f_item)

    # make sample folder
    if not os.path.exists(args.prefix + "_STAR-SEQR"):
        os.makedirs(args.prefix + "_STAR-SEQR")
    os.chdir(args.prefix + "_STAR-SEQR")

    # Do alignment if fastqs
    if args.fastq1:
        su.star_funcs.run_star(args.fastq1, args.fastq2, args)
    elif args.star_jxns:  # symlink existing files into folder if provided
        su.common.force_symlink(args.star_jxns, args.prefix + ".Chimeric.out.junction")
        su.common.force_symlink(args.star_sam, args.prefix + ".Chimeric.out.sam")

    # import all jxns
    rawdf = su.core.import_starjxns(args.prefix + ".Chimeric.out.junction", args.keep_dups)

    # get jxns only
    jxns = rawdf[rawdf['jxntype'] >= 0].reset_index()  # junctions can be either 0, 1, 2

    # Prepare Annotation
    global gtree # necessary to make global for multiprocessing at the moment
    gtree = su.gtf_convert.gtf2tree(args.gtf)

    if args.nucleic_type == "RNA":
        # start output files
        breakpoints_fh = open(args.prefix + "_STAR-SEQR_breakpoints.txt", 'w')
        breakpoint_cols = ["ann", "span_first", "jxn_left", "jxn_right",
                               "Fusion_Class", "splice_type", "breakpoint_left", "breakpoint_right",
                               "left_symbol", "right_symbol", "ann_format", "left_annot", "right_annot",
                               "dist", "assembly", "assembly_cross_disp", "primers", "name",
                               "span_homology_score", "jxn_homology_score", "overhang_diversity",
                               "minfrag20", "minfrag35", "disposition"]
        breakpoint_header = ["NAME", "NREAD_SPANS", "NREAD_JXNLEFT", "NREAD_JXNRIGHT",
                             "FUSION_CLASS", "SPLICE_TYPE", "BRKPT_LEFT", "BRKPT_RIGHT",
                             "LEFT_SYMBOL", "RIGHT_SYMBOL", "ANNOT_FORMAT", "LEFT_ANNOT", "RIGHT_ANNOT",
                             "DISTANCE", "ASSEMBLED_CONTIGS", "ASSEMBLY_CROSS_JXN","PRIMERS", "ID",
                             "SPAN_CROSSHOM_SCORE", "JXN_CROSSHOM_SCORE", "OVERHANG_DIVERSITY",
                             "MINFRAG20", "MINFRAG35", "DISPOSITION"]
        print('\t'.join(map(str, breakpoint_header)), file=breakpoints_fh)

        # stats dict
        stats_res = OrderedDict([('All_Breakpoints', 0), ('Candidate_Breakpoints', 0), ('Passing_Breakpoints', 0)])

        if len(jxns.index) == 0:
            logger.info("No junctions found in the input file")
            rna_closeout(args.prefix, stats_res, breakpoints_fh)
            sys.exit(0)

        # Order, Normalize and Aggregate
        logger.info("Ordering junctions")
        jxns['order'] = 1
        logger.info('Normalizing junctions')
        jxns = su.common.pandas_parallel(jxns, apply_normalize_jxns, args.threads)
        logger.info("Getting gene strand and flipping info as necessary")
        jxns = su.common.pandas_parallel(jxns, apply_jxn_strand, args.threads)
        jxns = su.common.pandas_parallel(jxns, apply_flip_func, args.threads)
        logger.info("Aggregating junctions")
        jxn_summary = su.core.count_jxns(jxns, args.nucleic_type)

        # write stats
        stats_res['All_Breakpoints'] = len(jxn_summary.index)
        logger.info('Total Breakpoints:' + str(stats_res['All_Breakpoints']))

        # Get discordant pairs
        logger.info('Getting pair info')
        # break the rawdf into chromosome specific files for quicker lookups
        dd = {}
        for chrom in set(rawdf['chrom1'].unique()) | set(rawdf['chrom2'].unique()):
            dd[chrom] = rawdf[(rawdf['chrom1'] == chrom) & (rawdf['jxntype'] == -1)]
        jxn_filt = su.common.pandas_parallel(jxn_summary, apply_pairs_func, args.threads, dd)

        # Require at least two reads for processing in order to reduce run time
        logger.info('Filtering junctions')
        before_remove = len(jxn_filt.index)
        jxn_filt = jxn_filt[(jxn_filt["spans"].astype(int) + jxn_filt["jxn_counts"].astype(int)) >= 2]
        logger.info("Number of candidates removed due to total read support less than 2: " + str(before_remove - len(jxn_filt.index)))

        if len(jxn_filt.index) >= 1:
            # Get Annotation info for each junction
            logger.info('Annotating junctions')
            jxn_filt = su.common.pandas_parallel(jxn_filt, wrap_annotate_junctions, args.threads)

            # determine if junction follows canonical splicing at exon junction
            logger.info('Getting junction info')
            jxn_filt = su.common.pandas_parallel(jxn_filt, wrap_jxn_info, args.threads)

            # subset to ROI using bed file if it exists
            if args.bed_file:
                logger.info('Subsetting junctions using the supplied bed file')
                before_remove = len(jxn_filt.index)
                targets_tree = su.common.bed_to_tree(args.bed_file)
                jxn_filt['subset'] = jxn_filt.apply(lambda x: su.common.subset_bed_func(x['name'], targets_tree, sub_style=args.subset_type), axis=1)
                jxn_filt = jxn_filt[jxn_filt['subset'] >= 1]
                logger.info("Number of candidates removed due to bed filter: " + str(before_remove - len(jxn_filt.index)))

        if len(jxn_filt.index) >= 1:
            # remove internal gene dups unless otherwise requested, also removes overlapping genes
            if not args.keep_gene_dups:
                before_remove = len(jxn_filt.index)
                jxn_filt = jxn_filt[jxn_filt['txintersection'].astype(str).str.len() < 3]  # [] counts as two
                logger.info("Number of candidates removed due to internal gene duplication filter: " + str(before_remove - len(jxn_filt.index)))

        if len(jxn_filt.index) >= 1:
            # remove novel genes unless otherwise requested
            # if not args.keep_novel:
            before_remove = len(jxn_filt.index)
            jxn_filt = jxn_filt[((jxn_filt['left_symbol'] != "NA") & (jxn_filt['right_symbol'] != "NA"))] # No Gene Symbol
            jxn_filt = jxn_filt[((jxn_filt['left_annot'].str.split(':', expand=True)[3] != "NA") & # No Exon number
                                (jxn_filt['right_annot'].str.split(':', expand=True)[3] != "NA"))]
            logger.info("Number of candidates removed due to novel gene filter: " + str(before_remove - len(jxn_filt.index)))

        if len(jxn_filt.index) >= 1:
            # remove mitochondria unless otherwise requested
            if not args.keep_mito:
                before_remove = len(jxn_filt.index)
                jxn_filt = jxn_filt[~jxn_filt['name'].str.contains("chrM|M")]
                logger.info("Number of candidates removed due to Mitochondria filter: " + str(before_remove - len(jxn_filt.index)))

        if len(jxn_filt.index) >= 1:
            # remove non-canonical jxns with less than 3 reads here to reduce run time
            jxn_filt = jxn_filt[~((jxn_filt['splice_type'] == "NON-CANONICAL_SPLICING") &
                                   ((jxn_filt["jxn_counts"].astype(int)) < 3))]

        if len(jxn_filt.index) >= 1:
            # combine all supporting reads together.
            jxn_filt['supporting_reads'] = jxn_filt['jxn_reads'] + ',' + jxn_filt['spanreads']

            stats_res['Candidate_Breakpoints'] = len(jxn_filt.index)
            logger.info('Candidate Breakpoints:' + str(stats_res['Candidate_Breakpoints']))

        # Process candidates
        if len(jxn_filt.index) >= 1:
            # Convert sam to bam
            su.common.sam_2_coord_bam(args.prefix + ".Chimeric.out.sam", args.prefix + ".Chimeric.out.bam", args.threads)
            su.common.check_file_exists(args.prefix + ".Chimeric.out.bam")

            # Gather unique read support
            logger.info("Getting read support.")
            su.common.make_new_dir('support')
            in_bam = args.prefix + ".Chimeric.out.bam"
            support_df = su.common.pandas_parallel(jxn_filt, apply_get_rna_support, args.threads, in_bam)
            finaldf = pd.merge(jxn_filt, support_df, how='inner', left_on="name", right_on="name", left_index=False,
                               right_index=True, sort=True, suffixes=('_x', '_y'), copy=True, indicator=False)
            # collapse read info for brevity but keep here in case useful later on
            finaldf['jxn_left'] = finaldf["jxnleft_for_first"] + finaldf["jxnleft_rev_first"]
            finaldf['jxn_right'] = finaldf["jxnright_for_first"] + finaldf["jxnright_rev_first"]
            finaldf['jxn_first'] = finaldf["jxnleft_for_first"] + finaldf["jxnleft_rev_first"] + \
                finaldf["jxnright_for_first"] + finaldf["jxnright_rev_first"]
            finaldf['jxn_second'] = finaldf["jxnleft_for_second"] + finaldf["jxnleft_rev_second"] + \
                finaldf["jxnright_for_second"] + finaldf["jxnright_rev_second"]
            finaldf['span_first'] = finaldf["spanleft_for_first"] + finaldf["spanleft_rev_first"] + \
                finaldf["spanright_for_first"] + finaldf["spanright_rev_first"]
            finaldf['span_second'] = finaldf["spanleft_for_second"] + finaldf["spanleft_rev_second"] + \
                finaldf["spanright_for_second"] + finaldf["spanright_rev_second"]
            finaldf['spans_disc'] = finaldf['span_first']

            # Get all overlapping transcript seqs into one fasta per side
            su.common.pandas_parallel(finaldf, wrap_exons2seq, args.threads, args.fasta)  # writes fasta for each transcript per side in support folder

            # get homology mapping scores
            logger.info("Getting read homology mapping scores")
            finaldf = su.common.pandas_parallel(finaldf, apply_get_cross_homology, args.threads)

            # get overhang read diversity
            logger.info("Getting overhang read diversity")
            finaldf = su.common.pandas_parallel(finaldf, apply_get_diversity, args.threads)

            # get min frag lengths for anchor/overhang
            logger.info("Getting min frag length support")
            finaldf = su.common.pandas_parallel(finaldf, apply_get_minfrag_length, args.threads)

            # get assembly seq and confirm breakpoint
            logger.info("doing assembly")
            finaldf = su.common.pandas_parallel(finaldf, apply_get_assembly_info, args.threads, args.as_type)
            finaldf['assembly_cross_disp'] = finaldf['assembly_cross_fusions'].apply(lambda x: True if len(str(x)) > 1 else False)

            # Generate Primers
            logger.info("Generating primers using indexed fasta")
            finaldf = su.common.pandas_parallel(finaldf, apply_primers_func, args.threads)

            # Get breakpoint locations
            logger.info("Getting normalized breakpoint locations")
            finaldf['breakpoint_left'], finaldf['breakpoint_right'] = zip(*finaldf.apply(lambda x: su.core.get_fusion_locations(x['name']), axis=1))

            # get homology mapping scores
            logger.info("Getting read homology mapping scores")
            finaldf = su.common.pandas_parallel(finaldf, apply_get_fusion_class, args.threads)

            # Get annotations for known fusions
            logger.info("Getting annotation")
            chimerdb3_path = su.common.find_resource('ChimerDB3.0_ChimerSeq.bedpe')
            fuca_path = su.common.find_resource('FusionCancer.bedpe')
            chimerdb3 = pd.read_csv(chimerdb3_path, sep="\t", header=None, low_memory=False, engine='c')
            chimerdb3.columns = ['c1', 's1', 'e1', 'c2', 's2', 'e2', 'Fusion_pair', 'Counts',
                                 'strand1', 'strand2', 'Source', 'Cancer', 'Features']
            fuca = pd.read_csv(fuca_path, sep="\t", header=None, low_memory=False, engine='c')
            fuca.columns = ['c1', 's1', 'e1', 'c2', 's2', 'e2', 'Database_ID', 'Rate',
                            'strand1', 'strand2', 'Methods', 'Cancers', 'Fusion_name']
            finaldf = su.common.pandas_parallel(finaldf, apply_get_annot_db, args.threads, chimerdb3, fuca)

            # TODO: probabilistic module to assign quality score

            # HARD FILTERING - Change this once a probabilistic module is ready.
            # Hard filter on read counts after accounting for transcript info.
            finaldf['filter_minreads'] = (((finaldf["jxn_first"] >= 2) |  # most robust cases
                                   ((finaldf["span_first"] >= 1) & (finaldf["jxn_left"] >= 1)) |  # read diversity
                                   ((finaldf["span_first"] >= 1) & (finaldf["jxn_right"] >= 1)) |
                                   ((finaldf["jxn_right"] >= 1) & (finaldf["jxn_left"] >= 1))))
            finaldf['filter_minreads'].replace(to_replace=[False], value='minreads', inplace=True, method=None)

            # Hard filter on homology for discordant pairs and jxn. Junctions sequences are usually smaller. Consider a ratio of score to read len?
            finaldf['filter_homology'] = ((finaldf['span_homology_score'] < .40) &
                                          (finaldf['jxn_homology_score'] < .40))
            finaldf['filter_homology'].replace(to_replace=[False], value='homology', inplace=True, method=None)


            # Hard filter on unique overhangs. Requre at least 20% of overhangs to be unique
            finaldf['filter_diversity'] = ((finaldf['overhang_diversity'] >= finaldf['jxn_first'] * .2))
            finaldf['filter_diversity'].replace(to_replace=[False], value='diversity', inplace=True, method=None)

            # Hard filter to require non-canonical splicing events to have greater read support and at least 10% with minfrag35
            noncan_mask = finaldf[finaldf['splice_type'] == "NON-CANONICAL_SPLICING"]
            finaldf['filter_noncanonical'] = ((noncan_mask['jxn_first'] >= 5) &
                                              (noncan_mask['minfrag35'] >= noncan_mask['jxn_first'] * .1))
            finaldf['filter_noncanonical'].replace(to_replace=[False], value='noncanonical_support', inplace=True, method=None)

            # Hard filter to require at least 10% of reads to pass minfrag20 if span reads == 0
            nospan_mask = finaldf[finaldf["span_first"] == 0]
            finaldf['filter_nospanminfrag'] = (nospan_mask['minfrag20'] >= nospan_mask['jxn_first'] * .1)
            finaldf['filter_nospanminfrag'].replace(to_replace=[False], value='nospan_minfrag', inplace=True, method=None)

            # Hard filter to require at least 10% of reads to pass minfrag20 if jxn reads > 10
            highjxn_mask = finaldf[finaldf["jxn_first"] >= 10]
            finaldf['filter_minfrag'] = (highjxn_mask['minfrag20'] >= highjxn_mask['jxn_first'] * .1)
            finaldf['filter_minfrag'].replace(to_replace=[False], value='minfrag', inplace=True, method=None)

            # Get the final disposition of filtering
            finaldf['filter_all'] = finaldf[['filter_minreads', 'filter_homology', 'filter_diversity', 'filter_noncanonical', 'filter_nospanminfrag', 'filter_minfrag']].values.tolist()
            finaldf['disposition'] = finaldf['filter_all'].apply(lambda x: ','.join(x for x in list(map(str, x)) if x not in ['True', 'nan']))
            finaldf['disposition'].replace('', 'PASS', inplace=True)

            # write candidates and info to file
            candid_fh = open(args.prefix + "_STAR-SEQR_candidates.txt", 'w')
            print('\t'.join(map(str, breakpoint_header)), file=candid_fh)
            finaldf.to_csv(path_or_buf=candid_fh, header=False, sep="\t", columns=breakpoint_cols, mode='w', index=False)
            candid_fh.close()

            resultsdf = finaldf[finaldf['disposition'] == 'PASS'].sort_values(['jxn_first', "span_first"], ascending=[False, False])

            # Write passing fusions to file
            resultsdf.to_csv(path_or_buf=breakpoints_fh, header=False, sep="\t",
                             columns=breakpoint_cols, mode='w', index=False)

            # Log Stats
            stats_res['Passing_Breakpoints'] = len(resultsdf.index)
            logger.info('Passing Breakpoints:' + str(stats_res['Passing_Breakpoints']))

            # closeout
            rna_closeout(args.prefix, stats_res, breakpoints_fh)

        else: # there were no candidates to process but still want result files produced
            # closeout
            logger.info('No candidates left to process.')
            rna_closeout(args.prefix, stats_res, breakpoints_fh)


    elif args.nucleic_type == "DNA":
        # start output files
        stats_fh = open(args.prefix + "_STAR-SEQR.stats", 'w')
        breakpoints_fh = open(args.prefix + "_STAR-SEQR_breakpoints.txt", 'w')
        breakpoint_cols = ["ann", "svtype", "breakpoint_left", "breakpoint_right", "dist", "spans_disc", "jxn_first", "jxn_second", "name"]
        breakpoint_header = ["NAME", "SVTYPE", "BRKPT_LEFT", "BRKPT_RIGHT", "DISTANCE", "NREAD_SPANS", "NREAD_JXNLEFT", "NREAD_JXNRIGHT", "ID"]
        print(*breakpoint_header, sep='\t', file=breakpoints_fh)

        # stats dict
        stats_res = {'All_Breakpoints': 0, 'Candidate_Breakpoints': 0, 'Passing_Breakpoints': 0}

        if len(jxns.index) == 0:
            logger.info("No junctions found in the input file")
            su.sv2bedpe.process(jxns, args)
            sys.exit(0)

        # Order, Normalize and Aggregate
        logger.info("Ordering junctions")
        jxns = su.common.pandas_parallel(jxns, apply_choose_order, args.threads)
        logger.info('Normalizing junctions')
        jxns = su.common.pandas_parallel(jxns, apply_normalize_jxns, args.threads)
        logger.info("Aggregating junctions")
        jxn_summary = su.core.count_jxns(jxns, nucleic_type = "DNA")
        # get distance
        jxn_summary['dist'] = jxn_summary.apply(lambda x: su.core.get_distance(x['name']), axis=1)

        # subset to bed if specified
        if args.bed_file:
            start_bed = time.time()
            logger.info('Subsetting junctions using the supplied bed file')
            targets_tree = su.common.bed_to_tree(bed_file)
            jxn_summary['subset'] = jxn_summary.apply(lambda x: su.common.subset_bed_func(x['name'], targets_tree), axis=1)
            jxn_summary = jxn_summary[jxn_summary['subset'] >= 1]
            logger.info("Time to subset junction from bed took %g seconds" % (time.time() - start_bed))

        # print stats
        stats_res['All_Breakpoints'] = len(jxn_summary.index)
        logger.info('Total Breakpoints:' + str(stats_res['All_Breakpoints']))

        # Filter on distance
        logger.info('Filtering junctions based on distance')
        jxn_summary = jxn_summary[(jxn_summary['dist'] >= args.dist) |
                                  (pd.isnull(jxn_summary['dist']))]
        logger.info('Junctions passing distance filter:' + str(len(jxn_summary.index)))

        # Filter on junctions
        logger.info('Filtering junctions based on number of junction reads')
        jxn_filt = jxn_summary[(jxn_summary["jxnleft"] >= args.jxn_reads) &
                               (jxn_summary["jxnright"] >= args.jxn_reads)].sort_values("jxnleft", ascending=False)
        logger.info('Junctions passing junction read filter:' + str(len(jxn_filt.index)))

        if len(jxn_filt.index) >= 1:
            # Annotate genes
            jxn_filt['genesleft'], jxn_filt['genesright'] = zip(*jxn_filt.apply(lambda x: su.annotate_sv.get_jxn_genes(x['name'], gtree), axis=1))
            jxn_filt['common'] = [list(set(a).intersection(set(b))) for a, b in zip(jxn_filt.genesleft, jxn_filt.genesright)]

            # Get gene info and remove internal gene dups
            if not args.keep_gene_dups:
                before_remove = len(jxn_filt.index)
                jxn_filt = jxn_filt[jxn_filt['common'].astype(str).str.len() < 3] # [] counts as two
                logger.info("Number of candidates removed due to internal gene duplication filter: " + str(before_remove - len(jxn_filt.index)))

        if len(jxn_filt.index) >= 1:
            # Get paired discordant spanning reads supporting junctions and filter
            logger.info('Getting pair info')
            dd = {}
            for chrom in set(rawdf['chrom1'].unique()) | set(rawdf['chrom2'].unique()):
                dd[chrom] = rawdf[(rawdf['chrom1'] == chrom) & (rawdf['jxntype'] == -1)]
            jxn_filt = su.common.pandas_parallel(jxn_filt, apply_pairs_func, args.threads, dd)
            logger.info('Filtering junctions based on pairs')
            jxn_filt = jxn_filt[(jxn_filt['spans'] >= args.span_reads)]

            # combine all supporting reads together.
            jxn_filt['supporting_reads'] = jxn_filt['spanreads'] + ',' + jxn_filt['jxnreadsleft'] + ',' + jxn_filt['jxnreadsright']

            # Log Stats
            stats_res['Candidate_Breakpoints'] = len(jxn_filt.index)
            logger.info('Candidate Breakpoints:' + str(stats_res['Candidate_Breakpoints']))

        if len(jxn_filt.index) >= 1:
            # skip the big functions for now
            finaldf = jxn_filt

            finaldf['jxn_first'] = finaldf['jxnleft']
            finaldf['jxn_second'] = finaldf['jxnright']
            finaldf['spans_disc'] = finaldf['spans']

            su.common.sam_2_coord_bam(args.prefix + ".Chimeric.out.sam", args.prefix + ".Chimeric.out.bam", args.threads)
            su.common.check_file_exists(args.prefix + ".Chimeric.out.bam")

            # Get Annotation

            finaldf['ann'] = finaldf['genesleft'].apply(lambda x: str(x[0])) + "--" + finaldf['genesright'].apply(lambda x: str(x[0]))

            # Get Breakpoint type
            finaldf['svtype'] = finaldf.apply(lambda x: su.core.get_svtype_func(x['name']), axis=1)

            # Get breakpoint locations
            finaldf['breakpoint_left'], finaldf['breakpoint_right'] = zip(*finaldf.apply(lambda x: su.core.get_sv_locations(x['name']), axis=1))
            finaldf.to_csv(path_or_buf="STAR-SEQR_candidate_info.txt", header=True, sep="\t", mode='w', index=False)

            finaldf.sort_values(['jxnleft', "spans"], ascending=[False, False], inplace=True)
            finaldf.to_csv(path_or_buf=breakpoints_fh, header=False, sep="\t",
                           columns=breakpoint_cols, mode='w', index=False)
            breakpoints_fh.close()

            # Make bedpe and VCF
            su.sv2bedpe.process(finaldf, args)

            # Log Stats
            stats_res['Passing_Breakpoints'] = len(finaldf.index)
            logger.info('Passing Breakpoints:' + str(stats_res['Passing_Breakpoints']))
        else:
            logger.info("No candidate junctions identified.")
            # Make bedpe and VCF, write headers only
            su.sv2bedpe.process(jxn_filt, args)

        # Write stats to file
        for key, value in stats_res.items():
            stats_fh.write(key + "\t" + str(value) + "\n")

    # Finish
    logger.info("Program took  %g seconds" % (time.time() - start))


if __name__ == "__main__":
    sys.exit(main())



import sys
import gzip
from operator import itemgetter
import argparse
from svtools.vcf.file import Vcf
from svtools.bedpe import Bedpe

def get_var_string(bedpe, cohort_name):
    if len(bedpe.cohort_vars) > 0:
        bedpe.set_info(cohort_name + '_AF', ','.join([value for (key, value) in sorted(bedpe.cohort_vars.items(), key=itemgetter(1),reverse=True)]))
        bedpe.set_info(cohort_name + '_VarID', ','.join([key for (key, value) in sorted(bedpe.cohort_vars.items(), key=itemgetter(1),reverse=True)]))
    else:
        bedpe.set_info(cohort_name + '_AF', str(0))
        bedpe.set_info(cohort_name + '_VarID', 'NONE')
    return str(bedpe)

def add(a_bedpe, b_bedpe, max_distance):
    if a_bedpe.svtype ==  b_bedpe.svtype:
        if (a_bedpe.o1 != b_bedpe.o1
            or a_bedpe.o2 != b_bedpe.o2):
            return False

        if (a_bedpe.c1 != b_bedpe.c1
            or a_bedpe.s1 - max_distance > b_bedpe.e1
            or a_bedpe.e1 + max_distance < b_bedpe.s1):
            return False

        if (a_bedpe.c2 != b_bedpe.c2
            or a_bedpe.s2 - max_distance > b_bedpe.e2
            or a_bedpe.e2 + max_distance < b_bedpe.s2):
            return False
        else:
            a_bedpe.cohort_vars[b_bedpe.name] = b_bedpe.af
            return True
    else:
        return False

def varLookup(aFile, bFile, bedpe_out, max_distance, pass_prefix, cohort_name):
    # FIXME The following code is heavily duplicated with vcftobedpe and bedpetovcf. Harmonize!!!
    bList = list()
    headerObj=Vcf() #co-opt the VCF header object
    if cohort_name is None:
        cohort_name=str(str(bFile).split('/')[-1])
        
    if bFile == "stdin":
        bData = sys.stdin
    elif bFile.endswith('.gz'):
        bData = gzip.open(bFile, 'rb')
    else:
        bData = open(bFile, 'r')
    for bLine in bData:
        if bLine.startswith(pass_prefix):
            continue
        bentry = Bedpe(bLine.rstrip().split('\t'))
        if bentry.af is None:
            sys.stderr.write('No allele frequency for variant found in -b file. This tool requires allele frequency information to function. Please add with svtools afreq and rerun\n')
            sys.exit(1)
        bList.append(bentry)
    
    if aFile == "stdin":
        aData = sys.stdin
    elif aFile.endswith('.gz'):
        aData = gzip.open(aFile, 'rb')
    else:
        aData = open(aFile, 'r')
    in_header=True    
    header_lines = []
    sample_list = None
    for aLine in aData:
        if pass_prefix is not None and aLine.startswith(pass_prefix):
            if aLine[0] == '#' and aLine[1] != '#':
                sample_list = aLine.rstrip().split('\t', 14)[-1]
            else:
                header_lines.append(aLine)
            continue
        else:
            if in_header == True:
                headerObj.add_header(header_lines)
                headerObj.add_info(cohort_name + '_AF', '.', 'Float', 'Allele frequency(ies) for matching variants found in the ' + cohort_name + ' vcf' + ' (' + str(str(bFile).split('/')[-1]) + ')' )
                headerObj.add_info(cohort_name + '_VarID', '.', 'Integer', 'List of Variant ID(s) for matching variants found in the ' + cohort_name + ' vcf' + ' (' + str(str(bFile).split('/')[-1]) + ')' )

                header = headerObj.get_header()
                bedpe_out.write(header[:header.rfind('\n')] + '\n')                
                if len(sample_list) > 0:
                    bedpe_out.write('\t'.join(['#CHROM_A',
                                               'START_A',
                                               'END_A',
                                               'CHROM_B',
                                               'START_B',
                                               'END_B',
                                               'ID',
                                               'QUAL',
                                               'STRAND_A',
                                               'STRAND_B',
                                               'TYPE',
                                               'FILTER',
                                               'INFO_A','INFO_B',
                                               sample_list]
                                             ) + '\n')
                else:
                    bedpe_out.write('\t'.join(['#CHROM_A',
                                               'START_A',
                                               'END_A',
                                               'CHROM_B',
                                               'START_B',
                                               'END_B',
                                               'ID',
                                               'QUAL',
                                               'STRAND_A',
                                               'STRAND_B',
                                               'TYPE',
                                               'FILTER',
                                               'INFO_A','INFO_B']
                                              ) + '\n')
                in_header=False
            a = Bedpe(aLine.rstrip().split('\t'))
            if a.af is None:
                sys.stderr.write('No allele frequency for variant found in -a file. This tool requires allele frequency information to function. Please add with svtools afreq and rerun\n')
                sys.exit(1)
            for b in bList:
                add(a,b,max_distance)
            bedpe_out.write(get_var_string(a, cohort_name) + '\n')

def description():
    return 'look for variants common between two BEDPE files'

def add_arguments_to_parser(parser):
    parser.add_argument('-d', '--distance', metavar='<INT>', type=int, dest='max_distance', default=50, help='max separation distance (bp) of adjacent loci between bedpe files [50]')
    parser.add_argument("-a", "--aFile", dest="aFile", metavar='<BEDPE>', help="pruned, merged BEDPE (A file) or standard input (-a stdin).")
    parser.add_argument("-b", "--bFile", dest="bFile", metavar='<BEDPE>', help="pruned merged BEDPE (B file) (-b stdin). For pruning use svtools prune")
    parser.add_argument("-c", "--cohort", dest='cohort_name', metavar='<STRING>', default=None, help="cohort name to add information of matching variants (default:bFile)")                    
    parser.add_argument('-o', '--output', type=argparse.FileType('w'), metavar='<BEDPE>', default=sys.stdout, help='output BEDPE to write (default: stdout)')
    parser.set_defaults(entry_point=run_from_args)

def command_parser():
    parser = argparse.ArgumentParser(description=description())
    add_arguments_to_parser(parser)
    return parser

def run_from_args(args):
    pass_prefix = "#"
    if args.aFile == None:
        if sys.stdin.isatty():
            sys.stderr.write('Please stream in input to this command or specify the file to read\n')
            sys.exit(1)
        else:
            args.aFile = sys.stdin

    try:
        varLookup(args.aFile, args.bFile, args.output, args.max_distance, pass_prefix, args.cohort_name)
    except IOError as err:
        sys.stderr.write("IOError " + str(err) + "\n");

if __name__ == '__main__':
    parser = command_parser()
    args = parser.parse_args()
    sys.exit(args.entry_point(args))



class Cluster(object):
    '''
    Stores information about overlapping SV and tracks the "best" one of the group
    '''
    def __init__(self):
        self.elements = [None]
        self.chrom_a = None
        self.min_a = float("inf")
        self.max_a = 0
        self.chrom_b = None
        self.min_b = float("inf")
        self.max_b = 0
        self.size = 0
        self.strand_a = ''
        self.strand_b = ''
        self.sv_event = ''
        self.filter = 0

    def can_add(self, bedpe, max_distance):
        '''
        Check whether a bedpe object is addable to this cluster given the max_distance
        '''
        if self.size == 0:
            return True
        if self.sv_event ==  bedpe.svtype:
            if (self.strand_a != bedpe.o1 or
                    self.strand_b != bedpe.o2):
                return False

            if (self.chrom_a != bedpe.c1 or
                    self.min_a - max_distance > bedpe.e1 or
                    self.max_a + max_distance < bedpe.s1):
                return False

            if (self.chrom_b != bedpe.c2 or
                    self.min_b - max_distance > bedpe.e2 or
                    self.max_b + max_distance < bedpe.s2):
                return False
            else:
                return True
        else:
            return False

    def add(self, bedpe, eval_param):
        '''
        Add a new Bedpe object to this cluster
        '''
        if eval_param is None or eval_param.lower() == 'af':
            if  bedpe.af != '.' and bedpe.af > self.filter:
                #First node represents best variant
                self.elements[0] = bedpe
        self.size += 1
        self.sv_event=bedpe.svtype
        self.filter = max(self.filter,bedpe.af)
        self.chrom_a = bedpe.c1
        self.min_a = min(self.min_a, bedpe.s1)
        self.max_a = max(self.max_a, bedpe.e1)
        self.chrom_b = bedpe.c2
        self.min_b = min(self.min_b, bedpe.s2)
        self.max_b = max(self.max_b, bedpe.e2)
        self.strand_a = bedpe.o1
        self.strand_b = bedpe.o2

    def get_cluster_string(self):
        '''
        The string for the best bedpe object for this cluster
        '''
        if self.elements[0] is not None:
            return str(self.elements[0])
        else:
            raise ValueError("Cluster doesn't contain a valid BEDPE line")



import argparse, sys
import gzip

MAX_SPLIT = 9

class Vcfpaste(object):
    def __init__(self, vcf_list, master=None, sum_quals=None):
        self.vcf_list = vcf_list
        self.master = master
        self.sum_quals = sum_quals

    def execute(self, output_handle=sys.stdout):
        try:
            self.read_filenames()
            self.open_files()
            self.write_header(output_handle)
            self.write_variants(output_handle)
        finally:
            self.close_files()

    def read_filenames(self):
        self.vcf_file_names = []
        with open(self.vcf_list, 'r') as vcf_list_file:
            for line in vcf_list_file:
                path = line.rstrip()
                self.vcf_file_names.append(path)
        if self.master == None:
            self.master = self.vcf_file_names[0]
        self.vcf_file_names.insert(0, self.master)

    def open_files(self):
        self.vcf_files = []
        # parse the vcf files to paste
        for path in self.vcf_file_names:
            if path.endswith('.gz'):
                self.vcf_files.append(gzip.open(path, 'rb'))
            else:
                self.vcf_files.append(open(path, 'r'))
    
    def write_header(self, output_handle=sys.stdout):
        master = self.vcf_files[0]
        while 1:
            master_line = master.readline()
            if not master_line:
                break
            if master_line[:2] != '##':
                break
            output_handle.write(master_line)
        out_v = master_line.rstrip().split('\t', MAX_SPLIT)[:(MAX_SPLIT - 1)] + ["FORMAT"]

        for vcf in self.vcf_files[1:]:
            while 1:
                l = vcf.readline()
                if not l:
                    break
                if l[:2] == '##':
                    continue
                if l[0] == '#':
                    out_v = out_v + l.rstrip().split('\t', MAX_SPLIT)[MAX_SPLIT:]
                    break
        output_handle.write('\t'.join(map(str, out_v)) + '\n')

    def write_variants(self, output_handle=sys.stdout):
        while 1:
            master_line = self.vcf_files[0].readline()
            if not master_line:
                break
            master_v = master_line.rstrip().split('\t', MAX_SPLIT)
            if len(master_v) < 8:
                sys.stderr.write('\nERROR: Master file {0} had less than 8 columns.\n'.format(self.vcf_files[0].name))
                exit(1)
            out_v = master_v[:8] # output array of fields
            qual = 0
            if out_v[5] != '.':
                qual = float(out_v[5])
            format = None # column 9, VCF format field.

            for vcf in self.vcf_files[1:]:
                line = vcf.readline()
                if not line:
                    sys.stderr.write('\nERROR: VCF files differ in length\n')
                    exit(1)
                line_v = line.rstrip().split('\t', MAX_SPLIT)
                if len(line_v) < 10:
                    sys.stderr.write('\nERROR: {0} had less than 10 columns. Only the master may be an 8 column VCF.\n'.format(vcf.name))
                    exit(1)

                # set FORMAT field as format in first VCF.
                # cannot extract this from master, since it may have
                # been altered in the processing of the VCFs.
                if format is None:
                    format = line_v[8]
                    out_v.append(format)

                if line_v[5] != '.':
                    qual += float(line_v[5])
                out_v = out_v + line_v[9:]
            if self.sum_quals:
                out_v[5] = qual
            output_handle.write( '\t'.join(map(str, out_v)) + '\n')

    def close_files(self):
        for f in self.vcf_files:
            f.close()

def description():
    return 'paste VCFs from multiple samples'

def epilog():
    return '''VCF files may be gzipped. If the -m argument is omitted then the first file in the list of files in --vcf-list is treated as the master.'''

def add_arguments_to_parser(parser):
    parser.add_argument('-f', '--vcf-list', metavar='<FILE>', required=True, help='file containing a line-delimited list of VCF files to paste (required)')
    parser.add_argument('-m', '--master', metavar='<VCF>', default=None, help='VCF file to set first 8 columns of variant info (otherwise first file in --vcf-list)')
    parser.add_argument('-q', '--sum-quals', required=False, action='store_true', help='sum QUAL scores of input VCFs as output QUAL score')
    parser.set_defaults(entry_point=run_from_args)

def command_parser():
    parser = argparse.ArgumentParser(description=description(), epilog=epilog())
    add_arguments_to_parser(parser)
    return parser

def run_from_args(args):
    paster = Vcfpaste(args.vcf_list, master=args.master, sum_quals=args.sum_quals)
    paster.execute()

# initialize the script
if __name__ == '__main__':
    parser = command_parser()
    args = parser.parse_args()
    sys.exit(args.entry_point(args))



import argparse, sys
from svtools.bedpe import Bedpe
from svtools.cluster import Cluster
import svtools.utils as su

class Pruner(object):
    def __init__(self, max_distance, eval_param):
        self.max_distance = max_distance
        self.eval_param = eval_param
        self.min_cluster_size = 1
        self.bedpe_lines = 0
        self.skipped_lines = 0
        self.emitted_lines = 0
        self.cluster_list = []

    def stats_report_string(self):
        return "{0} variants read\n{1} skipped\n{2} pruned\n{3} emitted\n".format(self.bedpe_lines,
            self.skipped_lines,
            self.bedpe_lines - self.skipped_lines - self.emitted_lines,
            self.emitted_lines)

    def cluster_bedpe(self, in_file, bedpe_out, is_sorted):
        # Locally alias instance variables
        max_distance = self.max_distance
        eval_param = self.eval_param

        in_header = True
        for line in in_file:
            if line.startswith('#') and in_header:
                bedpe_out.write(line)
                continue
            in_header = False
            self.bedpe_lines += 1
            bedpe = Bedpe(line.rstrip().split('\t'))
            if bedpe.af is None:
                sys.stderr.write('No allele frequency for variant found. This tool requires allele frequency information to function. Please add with svtools afreq and rerun\n')
                sys.exit(1)
            if bedpe.af == '.':
                self.skipped_lines += 1
                continue
            matched_clusters = []
            for cluster in self.cluster_list:
                if cluster.can_add(bedpe, max_distance):
                    cluster.add(bedpe, eval_param)
                    matched_clusters.append(cluster)
            if not matched_clusters:
                new_cluster = Cluster()
                new_cluster.add(bedpe, eval_param)
                self.cluster_list.append(new_cluster)
            else:
                if len(matched_clusters) > 1:
                    i = 0
                    matched_cluster_pruned = False
                    while i < (len(matched_clusters) - 1):
                        j = i + 1
                        while j < len(matched_clusters):
                            if matched_clusters[i].can_add(matched_clusters[j].elements[0], max_distance):
                                matched_clusters[i].add(matched_clusters[j].elements[0], eval_param)
                                matched_cluster_pruned = True
                                del matched_clusters[j]
                            j += 1
                        i += 1        
                    if matched_cluster_pruned:
                        self.cluster_list = [cluster for cluster in self.cluster_list if cluster not in matched_clusters]
            #prune and print eligible clusters
            if self.bedpe_lines % 1000 == 0 and is_sorted:
                self.cluster_list = self.prune(bedpe,
                                     False,
                                     bedpe_out)
    
        self.cluster_list = self.prune(None,
                             True,
                             bedpe_out)

        sys.stderr.write(self.stats_report_string())
        return

    def prune(self, bedpe, print_ineligible, bedpe_out):
        '''
        Prints and removes clusters from the cluster_list that are beyond the distance window
        '''
        min_cluster_size = self.min_cluster_size
        max_distance = self.max_distance
        new_cluster_list = []
        for cluster in self.cluster_list:
            # cluster is beyond updatable window:
            if (bedpe is None or 
                    cluster.chrom_a != bedpe.c1 or 
                    cluster.min_a - max_distance > bedpe.e1 or 
                    cluster.max_a + max_distance < bedpe.s1):
                
                # print the cluster if eligible
                if (cluster.size >= min_cluster_size or print_ineligible):
                    self.emitted_lines += 1
                    bedpe_out.write(cluster.get_cluster_string() + '\n')
                else:
                    new_cluster_list.append(cluster)
            # cluster is still within updatable window,
            # leave it in the cluster list
            else:
                new_cluster_list.append(cluster)
    
        return new_cluster_list

def description():
    return 'cluster and prune a BEDPE file by position based on allele frequency'

def epilog():
    return 'The input BEDPE file can be gzipped if it is specified explicitly.'

def add_arguments_to_parser(parser):
    parser.add_argument('-d', '--distance', metavar='<INT>', type=int, dest='max_distance', default=50, help='max separation distance (bp) of adjacent loci in cluster [50]')
    parser.add_argument('-e', '--eval-param', metavar='<STRING>', help='evaluating parameter for choosing best bedpe in a cluster(e.g. af=AlleleFrequency default:af)')
    parser.add_argument('-s', '--is-sorted', action='store_true', help='specify if an input file is sorted. Sort with svtools bedpesort. (default=False)')
    parser.add_argument('input', nargs='?', metavar='<BEDPE>', default=None, help='BEDPE file to read. If \'-\' or absent then defaults to stdin.')
    parser.add_argument('-o', '--output', metavar='<BEDPE>', type=argparse.FileType('w'), default=sys.stdout, help='output bedpe to write (default: stdout)')
    parser.set_defaults(entry_point=run_from_args)

def command_parser():
    parser = argparse.ArgumentParser(description=description(), epilog=epilog())
    add_arguments_to_parser(parser)
    return parser

def run_from_args(args):
    with su.InputStream(args.input) as stream:
        pruner = Pruner(args.max_distance, args.eval_param)
        pruner.cluster_bedpe(stream, args.output, args.is_sorted)

if __name__ == '__main__':
    parser = command_parser()
    args = parser.parse_args()
    sys.exit(args.entry_point(args))



#!/usr/bin/env python
# encoding: utf-8

from __future__ import (absolute_import, division, print_function)
import logging
import re

logger = logging.getLogger('STAR-SEQR')


def get_chimerdb(jxn, chimerdb3):
    chrom1, pos1, str1, chrom2, pos2, str2, repleft, repright = re.split(':', jxn)
    pos1 = int(pos1)
    pos2 = int(pos2)
    maxrep = max(10, int(max(int(repleft), int(repright))))
    chimerdb3.columns = ['c1', 's1', 'e1', 'c2', 's2', 'e2',
                  'Fusion', 'Counts', 'strand1', 'strand2',
                  'Source', 'Cancer', 'Features']
    # do a range query for both sides. Assume no switching head and tail
    subset_df = chimerdb3[((chimerdb3['c1']== chrom1) & ((chimerdb3['s1']>=pos1-maxrep) & (chimerdb3['s1']<=pos1+maxrep))) &
     ((chimerdb3['c2']== chrom2) & ((chimerdb3['s2']>=pos2-maxrep) & (chimerdb3['s2']<=pos2+maxrep)))]
    sum_count = ''
    if len(subset_df.index >= 1):
         sum_count = subset_df['Counts'].astype(int).sum()
    return (sum_count,
            ','.join(set(subset_df['Source'].astype(str))),
            ','.join(set(subset_df['Cancer'].astype(str))),
            ','.join(set(subset_df['Features'].astype(str))))


def get_fusioncancerdb(jxn, fuca):
    chrom1, pos1, str1, chrom2, pos2, str2, repleft, repright = re.split(':', jxn)
    pos1 = int(pos1)
    pos2 = int(pos2)
    maxrep = max(10, int(max(int(repleft), int(repright))))
    # do a range query for both sides. Assume no switching head and tail
    subset_df = fuca[((fuca['c1']== chrom1) & ((fuca['s1']>=pos1-maxrep) & (fuca['s1']<=pos1+maxrep))) &
     ((fuca['c2']== chrom2) & ((fuca['s2']>=pos2-maxrep) & (fuca['s2']<=pos2+maxrep)))]
    max_rate = ''
    if len(subset_df.index >= 1):
         max_rate = float("{0:.2f}".format(subset_df['Rate'].astype(float).max()))
    return (max_rate,
            ','.join(set(subset_df['Methods'].astype(str))),
            ','.join(set(subset_df['Cancers'].astype(str))))



from . import libssw
import six
from six.moves import range

__all__ = ["ScoreMatrix", "DNA_ScoreMatrix", "Aligner", "Alignment"]

def build_compliment_table():
    cmap = ((ord('G'), ord('C')), (ord('A'), ord('T')))
    _ctable = [chr(idx) for idx in range(0xff + 1)]
    case_delta = ord('a') - ord('A')
    for (base1, base2) in cmap:
        _ctable[base1] = chr(base2)
        _ctable[base2] = chr(base1)
        _ctable[base1 + case_delta] = chr(base2 + case_delta)
        _ctable[base2 + case_delta] = chr(base1 + case_delta)
    return str.join('', _ctable)
ComplimentTable = build_compliment_table()

class ScoreMatrix(object):
    def __init__(self, alphabet=None, match=2, mismatch=-2):
        self._match = match
        self._mismatch = mismatch
        self.alphabet = alphabet

    def get_mismatch(self):
        return self._mismatch
    def set_mismatch(self, val):
        self._mismatch = val
        self._init_matrix()
    mismatch = property(get_mismatch, set_mismatch)

    def get_match(self):
        return self._match
    def set_match(self, val):
        self._match = val
        self._init_matrix()
    match = property(get_match, set_match)

    def get_alphabet(self):
        return self._alphabet

    def set_alphabet(self, alphabet):
        self._alphabet = tuple(alphabet) if alphabet else tuple()
        self.symbol_map = {symbol.upper(): idx for (idx, symbol) in enumerate(self._alphabet)}
        self._init_matrix()
    alphabet = property(get_alphabet, set_alphabet)

    def _init_matrix(self):
        _matrix_type = libssw.matrix_type * (len(self.alphabet) ** 2)
        self._matrix = _matrix_type(*self.iter_matrix())

    def iter_matrix(self):
        for row_symbol in self.alphabet:
            for col_symbol in self.alphabet:
                yield self._get_score(row_symbol, col_symbol)

    def _get_score(self, symbol_1, symbol_2):
        if symbol_1.upper() == symbol_2.upper(): return self.match
        return self.mismatch

    def convert_sequence_to_ints(self, seq):
        seq = seq.upper()
        _seq_type = libssw.symbol_type * len(seq)
        _seq_instance = _seq_type()
        for (idx, symbol) in enumerate(seq):
            _seq_instance[idx] = self.symbol_map[symbol]
        return _seq_instance

class DNA_ScoreMatrix(ScoreMatrix):
    def __init__(self, alphabet='AGTCN', **kw):
        super(DNA_ScoreMatrix, self).__init__(alphabet=alphabet, **kw)

    def _get_score(self, symbol_1, symbol_2):
        if symbol_1.upper() == 'N' or symbol_2.upper() == 'N':
            return 0
        return super(DNA_ScoreMatrix, self)._get_score(symbol_1, symbol_2)

class Aligner(object):
    def __init__(self, reference=None, matrix=None, molecule="dna", gap_open=3, gap_extend=1):
        self.reference = reference
        self.matrix = matrix
        self.molecule = molecule
        if self.matrix == None and molecule != None:
            if molecule == "dna":
                self.matrix = DNA_ScoreMatrix()
            else:
                raise ValueError("Unrecognized molecule type '%s'" % molecule)
        self.gap_open = gap_open
        self.gap_extend = gap_extend
        if self.gap_open <= self.gap_extend:
            raise ValueError("gap_open must always be greater than gap_extend")

    def align(self, query='', reference=None, revcomp=True):
        # XXX: I really don't find this part of SSW useful, which
        # is why i broke alignment into two stages, so you can use 
        # the low level interface if you wish.
        filter_score = 0
        filter_distance = 0
        flags = 1
        mask_length = max(15, len(query) // 2)
        reference = reference if reference != None else self.reference
        res = self._align(query, reference, flags, filter_score, filter_distance, mask_length)
        if revcomp:
            query_rc = query[::-1].translate(ComplimentTable)
            res_rc = self._align(query_rc, reference, flags, filter_score, filter_distance, mask_length)
            if res_rc.score > res.score:
                res = res_rc
        return res
    
    def _align(self, query, reference, flags, filter_score, filter_distance, mask_length, score_size=2): 
        _query = self.matrix.convert_sequence_to_ints(query)
        _reference = self.matrix.convert_sequence_to_ints(reference)
        profile = libssw.ssw_profile_init(_query, len(query), self.matrix._matrix, len(self.matrix.alphabet), score_size)
        if self.gap_open <= self.gap_extend:
            raise ValueError("gap_open must always be greater than gap_extend")
        alignment = libssw.ssw_align_init(profile, _reference, len(_reference), self.gap_open, self.gap_extend, flags, filter_score, filter_distance, mask_length) 
        alignment_instance = Alignment(alignment, query, reference)
        libssw.ssw_profile_del(profile)
        libssw.ssw_align_del(alignment)
        return alignment_instance

class Alignment(object):
    def __init__ (self, alignment, query, reference):
        self.score = alignment.contents.score
        self.score2 = alignment.contents.score2
        self.reference = reference
        self.reference_begin = alignment.contents.ref_begin
        self.reference_end = alignment.contents.ref_end
        self.query = query
        self.query_begin = alignment.contents.query_begin
        self.query_end = alignment.contents.query_end
        self._cigar_string = [alignment.contents.cigar[idx] for idx in range(alignment.contents.cigarLen)]

    @property
    def iter_cigar(self):
        for val in self._cigar_string:
            op_len = libssw.cigar_int_to_len(val)
            op_char = libssw.cigar_int_to_op(val).decode("latin")
            yield (op_len, op_char)

    @property
    def cigar(self):
        cigar = ""
        if self.query_begin > 0:
            cigar += str(self.query_begin) + "S"
        cigar += str.join('', (str.join('', map(str, cstr)) for cstr in self.iter_cigar))
        end_len = len(self.query) - self.query_end - 1
        if end_len != 0:
            cigar += str(end_len) + "S"
        return cigar

    @property
    def alignment(self):
        r_index = 0
        q_index = 0
        r_seq = self.reference[self.reference_begin: self.reference_end + 1]
        q_seq = self.query[self.query_begin: self.query_end + 1]
        r_line = m_line = q_line = ''
        for (op_len, op_char) in self.iter_cigar:
            op_len = int(op_len)
            if op_char.upper() == 'M':
                ref_piece = r_seq[r_index: r_index + op_len]
                query_peace = q_seq[q_index: q_index + op_len]
                r_line += ref_piece
                q_line += query_peace
                match_seq = ''.join(['|' if r_base.upper() == q_base.upper() else '*' for (r_base, q_base) in zip(ref_piece, query_peace)]) # faster with "".join([list of str]) instead of +=
                    
                m_line += match_seq
                r_index += op_len
                q_index += op_len

            elif op_char.upper() == 'I':
                # insertion into reference
                r_line += '-' * op_len
                m_line += ' ' * op_len
                q_line += q_seq[q_index: q_index + op_len]
                #  only query index change
                q_index += op_len
            elif op_char.upper() == 'D':
                # deletion from reference
                r_line += r_seq[r_index: r_index + op_len]
                m_line += ' ' * op_len
                q_line += '-' * op_len
                #  only ref index change
                r_index += op_len
        return (r_line, m_line, q_line)

    # XXX: all of these count functions are ineffecient

    @property
    def match_count(self):
        return self.alignment[1].count('|')

    @property
    def mismatch_count(self):
        return self.alignment[1].count('*')

    @property
    def insertion_count(self):
        cnt = 0
        for (op_len, op_char) in self.iter_cigar:
            if op_char.upper() == 'I':
                cnt += op_len
        return cnt

    @property
    def deletion_count(self):
        cnt = 0
        for (op_len, op_char) in self.iter_cigar:
            if op_char.upper() == 'D':
                cnt += op_len
        return cnt

    def alignment_report(self, width=80, header=True):
        def window(lines, width):
            idx = 0
            while 1:
                res = []
                for line in lines:
                    res.append(line[idx:idx+width])
                if not any(res):
                    break
                yield (res, idx)
                idx += width

        margin_width = len(str(max(self.query_end, self.reference_end))) + 8
        rpt = ''
        if header:
            rpt += "Score = %s, Matches = %s, Mismatches = %s, Insertions = %s, Deletions = %s\n" % (self.score, self.match_count, self.mismatch_count, self.insertion_count, self.deletion_count)
            rpt += '\n'
        for (lines, offset) in window(self.alignment, width - margin_width):
            for (name, seq_offset, line) in zip(["ref", "", "query"], [self.reference_begin, None, self.query_begin], lines):
                if name:
                    line_offset = seq_offset + offset + 1
                    left_margin = "%s %s" % (name.ljust(5), line_offset)
                else:
                    left_margin = ""
                rpt += "%s%s\n" % (left_margin.ljust(margin_width), line)
            rpt += '\n'
        return rpt



#!/usr/bin/env python

import argparse, sys, copy, gzip, time, math, re
import numpy as np
import pandas as pd
from scipy import stats
from collections import Counter, defaultdict, namedtuple
import statsmodels.formula.api as smf
from operator import itemgetter
import warnings
from svtools.vcf.file import Vcf
from svtools.vcf.genotype import Genotype
from svtools.vcf.variant import Variant
import svtools.utils as su


CN_rec = namedtuple ('CN_rec', 'var_id sample svtype svlen AF GT CN AB log_len log2r')



# http://stackoverflow.com/questions/8930370/where-can-i-find-mad-mean-absolute-deviation-in-scipy
def mad(arr):
    """ Median Absolute Deviation: a "Robust" version of standard deviation.
        Indices variabililty of the sample.
        https://en.wikipedia.org/wiki/Median_absolute_deviation 
    """
    arr = np.ma.array(arr).compressed() # should be faster to not use masked arrays.
    med = np.median(arr)
    return np.median(np.abs(arr - med))


def to_bnd_strings(var, fixed_gts):

    old_type = var.info['SVTYPE']
    old_id = var.var_id
    old_pos = var.pos
    old_end = var.info['END']
    old_ciend = var.info['CIEND']
    old_cipos = var.info['CIPOS']
    old_cipos95 = var.info['CIPOS95']
    old_ciend95 = var.info['CIEND95']


    #for both ends
    var.info['SVTYPE'] = 'BND'
    var.info['EVENT'] = old_id
    del var.info['SVLEN']
    del var.info['END']

    #var1
    var.var_id = old_id + "_1"
    var.info['MATEID'] = old_id + "_2"
    if old_type == 'DEL':
        var.alt = 'N[%s:%s[' % (var.chrom, old_end)
    else:
        var.alt = ']%s:%s]N' % (var.chrom, old_end)
    var1=var.get_var_string(fixed_gts)

    #var2
    var.var_id = old_id + "_2"
    var.info['MATEID'] = old_id + "_1"
    var.info['CIPOS'] = old_ciend
    var.info['CIEND'] = old_cipos
    var.info['CIPOS95'] = old_ciend95
    var.info['CIEND95'] = old_cipos95
    var.pos = old_end
    var.info['SECONDARY'] = True
    if old_type == 'DEL':
        var.alt = ']%s:%s]N' % (var.chrom, old_pos)
    else:
        var.alt = 'N[%s:%s[' % (var.chrom, old_pos)
    var2=var.get_var_string(fixed_gts)
    return var1, var2


def reciprocal_overlap(a, b_list):
    overlap = 0
    b_aggregate = 0
    # catch divide by zero error
    if a[1] == a[0]:
        return 0

    # update the overlap and b_aggregate
    for b in b_list:
        b_aggregate += (b[1] - b[0])
        overlap += float(min(a[1], b[1]) - max(a[0], b[0]))

    # catch divide by zero error
    if b_aggregate == 0:
        return 0

    return min(overlap / (a[1] - a[0]), overlap / b_aggregate)

def collapse_bed_records(bed_list):

    bed_list_sorted = sorted(bed_list, key=itemgetter(1))
    collapsed_bed_list = []
    i = 0
    curr_rec = bed_list_sorted[i]
    while i < len(bed_list_sorted):
        # end at last element in list
        if i == len(bed_list_sorted) - 1:
            collapsed_bed_list.append(copy.copy(curr_rec))
            break

        # load next entry
        next_rec = bed_list_sorted[i + 1]
        # merge is overlap
        if curr_rec[1] >= next_rec[0]:
            curr_rec[1] = next_rec[1]
            i += 1
        # write out if no overlap
        else:
            collapsed_bed_list.append(copy.copy(curr_rec))
            i += 1
            curr_rec = bed_list_sorted[i]

    # print 'collapsed:', collapsed_bed_list
    return collapsed_bed_list

def annotation_intersect(var, ae_dict, threshold):
    best_frac_overlap = 0
    best_feature = ''
    slop = 0

    # dictionary with number of bases of overlap for each class
    class_overlap = {}
    
    # first check for reciprocal overlap
    if var.chrom in ae_dict:
        var_start = var.pos
        var_end = int(var.info['END'])
        i = 0
        while 1:
            # bail if end of dict
            if i >= len(ae_dict[var.chrom]):
                break
            feature = ae_dict[var.chrom][i]
            if feature[0] - slop < var_end:
                if feature[1] + slop > var_start:
                    try:
                        class_overlap[feature[2]].append(feature)
                    except KeyError:
                        class_overlap[feature[2]] = [feature]
            else:
                break
            i += 1

        # print class_overlap
        for me_class in class_overlap:
            class_overlap[me_class] = collapse_bed_records(class_overlap[me_class])
            frac_overlap = reciprocal_overlap([var_start, var_end], class_overlap[me_class])
            if frac_overlap > best_frac_overlap:
                best_frac_overlap = frac_overlap
                best_feature = me_class


        if best_frac_overlap >= threshold:
            return best_feature

    return None

def lowQuantile(xx):
    return np.percentile(xx,2.5)

def highQuantile(xx):
    return np.percentile(xx,97.5)

def lld(xx, mean, sd):
    ll = 1 / sd * math.exp(-(xx-mean) * (xx-mean) / (2*sd*sd))
    return ll

def calc_params(vcf_path):

    tSet = list()
    epsilon=0.1
    header=[]
    
    in_header = True
    vcf = Vcf()
    if vcf_path.endswith('.gz'):
        vcf_file = gzip.open(vcf_path, 'rb')
    else:
        vcf_file = open(vcf_path, 'r')

    for line in vcf_file:
        if in_header:
            if line[0] == '#':
                header.append(line)
                if line[1] != '#':
                    vcf_samples = line.rstrip().split('\t')[9:]
                    in_header = False
                    vcf.add_header(header)
                continue
        else:
            v = line.rstrip().split('\t')
            info = v[7].split(';')
            svtype = None
            for x in info:
                if x.startswith('SVTYPE='):
                    svtype = x.split('=')[1]
                    break

            if svtype not in ['DEL', 'DUP'] or v[0]=="X" or v[0]=="Y":
                continue

            var = Variant(v, vcf)
    
            for sample in vcf_samples:
                sample_genotype = var.genotype(sample)
                if sample_genotype.get_format('GT') != './.':
                    log2r = math.log((float(sample_genotype.get_format('CN'))+ epsilon)/2,2)  #to avoid log(0)
                    tSet.append(CN_rec(var.var_id, sample, var.info['SVTYPE'], abs(float(var.info['SVLEN'])), var.info['AF'],
                        sample_genotype.get_format('GT'),  sample_genotype.get_format('CN'), sample_genotype.get_format('AB'), math.log(abs(float(var.info['SVLEN']))), log2r))

    df=pd.DataFrame(tSet, columns=CN_rec._fields)
    #exclude from training data, DELs and DUPs with CN in the tails of the distribution
    df['q_low']=df.groupby(['sample', 'svtype', 'GT'])['log2r'].transform(lowQuantile)
    df['q_high']=df.groupby(['sample', 'svtype', 'GT'])['log2r'].transform(highQuantile)
    df=df[(df.log2r>=df.q_low) & (df.log2r<=df.q_high)]
    #df.to_csv('./train.csv')

    #adjust copy number for small deletions (<1kb), no strong relationship b/w cn and size for dups evident so far
    small_het_dels = df[(df.svtype=="DEL") & (df.GT=="0/1") & (df.svlen<1000) & (df.svlen>=50)]
    small_hom_dels = df[(df.svtype=="DEL") & (df.GT=="1/1") & (df.svlen<1000) & (df.svlen>=50)]
    het_del_mean=np.mean(df[(df.svlen>1000) & (df.GT=="0/1") & (df.svtype=="DEL")]['log2r'])
    hom_del_mean=np.mean(df[(df.svlen>1000) & (df.GT=="1/1") & (df.svtype=="DEL")]['log2r'])
    small_het_dels['offset']=small_het_dels['log2r']-het_del_mean
    small_hom_dels['offset']=small_hom_dels['log2r']-hom_del_mean
    
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        hom_del_fit=smf.ols('offset~log_len',small_hom_dels).fit()
        het_del_fit=smf.ols('offset~log_len',small_het_dels).fit()
        #print hom_del_fit.summary()
        #print het_del_fit.summary()
        small_hom_dels['log2r_adj'] = small_hom_dels['log2r'] - hom_del_fit.predict(small_hom_dels)
        small_het_dels['log2r_adj'] = small_het_dels['log2r'] - het_del_fit.predict(small_het_dels)

    small_dels=small_hom_dels.append(small_het_dels)
    small_dels=small_dels[['var_id', 'sample', 'svtype', 'svlen', 'AF', 'GT', 'CN', 'log_len', 'log2r', 'q_low', 'q_high', 'log2r_adj']]

    # dels of length<100 bp are excluded here
    df1=df[(df.svtype!="DEL") | (df.GT=="0/0") | (df.svlen>=1000)]
    df1['log2r_adj']=df1['log2r']
    df1=df1.append(small_dels)

    params=df1.groupby(['sample', 'svtype', 'GT'])['log2r_adj'].aggregate([np.mean,np.var, len]).reset_index()
    params=pd.pivot_table(params, index=['sample', 'svtype'], columns='GT', values=['mean', 'var', 'len']).reset_index()    
    params.columns=['sample', 'svtype', 'mean0', 'mean1', 'mean2', 'var0', 'var1', 'var2', 'len0', 'len1', 'len2']
    params['std_pooled']=np.sqrt((params['var0']*params['len0']+params['var1']*params['len1']+params['var2']*params['len2'])/(params['len0']+params['len1']+params['len2']))
    #params.to_csv('./params.csv')
    return (params, het_del_fit, hom_del_fit)


def rd_support_nb(temp, p_cnv):
    tr = pd.DataFrame({'p0' : [1.0, 0.1, 0.0], 'p1' : [0.0, 0.7, 0.25], 'p2' : [0.0, 0.2, 0.75], 'GT' : ["0/0", "0/1", "1/1"]})
    temp = pd.merge(temp, tr, on='GT', how='left')
    temp['p_mix'] = temp['lld0'] * temp['p0'] + temp['lld1'] * temp['p1'] + temp['lld2'] * temp['p2']
    return np.log(p_cnv)+np.sum(np.log(temp['p_mix'])) > np.log(1-p_cnv)+np.sum(np.log(temp['lld0']))
   

def has_rd_support_by_nb(test_set, het_del_fit, hom_del_fit, params, p_cnv = 0.5):
    svtype=test_set['svtype'][0]
    svlen=test_set['svlen'][0]
    log_len=test_set['log_len'][0]
    
    if svtype == 'DEL' and svlen<1000:
        params1=params[params.svtype=='DEL'].copy()
        if svlen<50:
            params1['log_len']=math.log(50)
        else:
            params1['log_len']=log_len
        params1['mean1_adj'] = params1['mean1'] + het_del_fit.predict(params1)
        params1['mean2_adj'] = params1['mean2'] + hom_del_fit.predict(params1)
    else:
        params1=params.copy()
        params1['mean1_adj'] = params1['mean1']
        params1['mean2_adj'] = params1['mean2']

    v0=test_set[test_set.GT=="0/0"]['log2r'].values
    v1=test_set[test_set.GT=="0/1"]['log2r'].values
    v2=test_set[test_set.GT=="1/1"]['log2r'].values

    if len(v0)>0:
        med0=np.median(v0)
    else:
        if len(v1)>0:
            med0=med1=np.median(v1)
        elif len(v2)>0:
            med0=med1=med2=np.median(v2)
        else:
            return False

    if len(v1)>0:
        med1=np.median(v1)
    else:
        med1=med0
    if len(v2)>0:
        med2=np.median(v2)
    else:
        med2=med1

    if svtype=='DEL' and ( med1>med0 or med2>med0 ):
        return False
    elif svtype=='DUP' and (med1<med0 or med2<med0):
        return False

    mm=pd.merge(test_set, params1, how='left')

    mm['lld0'] = mm.apply(lambda row:lld(row["log2r"], row["mean0"],row["std_pooled"]), axis=1)
    mm['lld1'] = mm.apply(lambda row:lld(row["log2r"], row["mean1_adj"],row["std_pooled"]), axis=1)
    mm['lld2'] = mm.apply(lambda row:lld(row["log2r"], row["mean2_adj"],row["std_pooled"]), axis=1)
   
    return  rd_support_nb(mm, p_cnv)


def load_df(var, exclude, sex):
    
    epsilon=0.1
    test_set = list()

    for s in var.sample_list:
        if s in exclude:
            continue
        cn = var.genotype(s).get_format('CN')
        if (var.chrom == 'X' or var.chrom == 'Y') and sex[s] == 1:
            cn=str(float(cn)*2)
        log2r = math.log((float(cn)+epsilon)/2, 2)  # to avoid log(0)
        test_set.append(CN_rec(var.var_id, s, var.info['SVTYPE'], abs(float(var.info['SVLEN'])), var.info['AF'],
             var.genotype(s).get_format('GT'),  cn , var.genotype(s).get_format('AB'), math.log(abs(float(var.info['SVLEN']))), log2r))

    test_set = pd.DataFrame(data = test_set, columns=CN_rec._fields)
    return test_set

# test for read depth support of low frequency variants
def has_low_freq_depth_support(test_set, mad_threshold=2, absolute_cn_diff=0.5):

    mad_quorum = 0.5 # this fraction of the pos. genotyped results must meet the mad_threshold
    
    hom_ref_cn=test_set[test_set.GT=="0/0"]['CN'].values.astype(float)
    hom_het_alt_cn=test_set[(test_set.GT=="0/1") | (test_set.GT=="1/1")]['CN'].values.astype(float)

    if len(hom_ref_cn) > 0:
        cn_median = np.median(hom_ref_cn)
        cn_mad = mad(hom_ref_cn)
    else:
        cn_median = None
        cn_mad = None

    # bail after writing out diagnostic info, if no ref samples or all ref samples
    if (len(hom_ref_cn) == 0 or len(hom_het_alt_cn) == 0):
        return False

    # tally up the pos. genotyped samples meeting the mad_threshold

    resid=hom_het_alt_cn-cn_median
    if test_set['svtype'][0]=='DEL':
        resid=-resid
    
    resid=resid[(resid > (cn_mad * mad_threshold) ) & (resid>absolute_cn_diff)]

    if float(len(resid))/len(hom_het_alt_cn)>mad_quorum:
        return True
    else:
        return False

# test whether variant has read depth support by regression
def has_high_freq_depth_support(df, slope_threshold, rsquared_threshold):
    
    rd = df[[ 'AB', 'CN']][df['AB']!='.'].values.astype(float)
    if len(np.unique(rd[0,:])) > 1 and len(np.unique(rd[1,:])) > 1:
        
        (slope, intercept, r_value, p_value, std_err) = stats.linregress(rd)
        if df['svtype'][0] == 'DEL':
            slope=-slope
        #sys.stderr.write(df['var_id'][0]+"\t"+str(slope)+"\t"+str(r_value)+"\n")

        if (slope < slope_threshold or r_value*r_value < rsquared_threshold):
            return False
        return True
    return False

def has_rd_support_by_ls(df, slope_threshold, rsquared_threshold, num_pos_samps, mad_threshold=2, absolute_cn_diff=0.5):

    min_pos_samps_for_regression=10
    if num_pos_samps>min_pos_samps_for_regression:
        return has_high_freq_depth_support(df, slope_threshold, rsquared_threshold)
    else:
        return has_low_freq_depth_support(df, mad_threshold, absolute_cn_diff)
    return False

def has_rd_support_hybrid(df, het_del_fit, hom_del_fit, params, p_cnv, slope_threshold, rsquared_threshold, num_pos_samps):
  
    hybrid_support=False
    nb_support=has_rd_support_by_nb(df, het_del_fit, hom_del_fit, params, p_cnv)
    ls_support=has_rd_support_by_ls(df, slope_threshold, rsquared_threshold, num_pos_samps)

    if nb_support and ls_support:
        hybrid_support=True
    elif nb_support and has_rd_support_by_ls(df, 2*slope_threshold, 2*rsquared_threshold, num_pos_samps, 2, 0.75):
        hybrid_support=True
    elif ls_support and has_rd_support_by_nb(df, het_del_fit, hom_del_fit, params, 0.2*p_cnv):
        hybrid_support=True
    return [ls_support, nb_support, hybrid_support]


# primary function
def sv_classify(vcf_in, vcf_out, gender_file, exclude_file, ae_dict, f_overlap, slope_threshold, rsquared_threshold, p_cnv, het_del_fit, hom_del_fit, params, diag_outfile, method):

    vcf = Vcf()
    header = []
    in_header = True
    sex = {}
    # read sample genders
    for line in gender_file:
        v = line.rstrip().split('\t')
        sex[v[0]] = int(v[1])

    exclude = []
    if exclude_file is not None:
        for line in exclude_file:
            exclude.append(line.rstrip())

    if diag_outfile is not None:
        outf=open(diag_outfile, 'w', 4096)
        outf.write("varid\torig_svtype\tsvlen\tnum_pos_samps\tnb_support\tls_support\thybrid_support\thas_rd_support\n")

    for line in vcf_in:
        if in_header:
            if line[0] == '#':
                header.append(line)
                continue
            else:
                in_header = False
                vcf.add_header(header)
                vcf_out.write(vcf.get_header() + '\n')

        v = line.rstrip().split('\t')
        info = v[7].split(';')
        svtype = None
        for x in info:
            if x.startswith('SVTYPE='):
                svtype = x.split('=')[1]
                break
        # bail if not DEL or DUP prior to reclassification
        if svtype not in ['DEL', 'DUP']:
            vcf_out.write(line)
            continue
        
        var = Variant(v, vcf)

        # check intersection with mobile elements
        if ae_dict is not None and var.info['SVTYPE'] in ['DEL']:
            ae = annotation_intersect(var, ae_dict, f_overlap)
            if ae is not None:
                if ae.startswith('SINE') or ae.startswith('LINE') or ae.split('|')[2].startswith('SVA'):
                    ae = 'ME:' + ae
                var.alt = '<DEL:%s>' % ae
                var.info['SVTYPE'] = 'MEI'
                vcf_out.write(var.get_var_string(True) + '\n')
                continue

        #count positively genotyped samples
        num_pos_samps = 0
        num_total_samps=len(var.sample_list)

        for s in var.sample_list:
            if var.genotype(s).get_format('GT') not in ["./.", "0/0"]:
                num_pos_samps += 1

        nb_support = False
        ls_support = False
        hybrid_support = False
        has_rd_support = False

        if num_pos_samps == 0:
            vcf_out.write(line)
        else:
            df=load_df(var, exclude, sex)
            if method=='large_sample':
                ls_support = has_rd_support_by_ls(df, slope_threshold, rsquared_threshold, num_pos_samps)
                has_rd_support=ls_support
            elif method=='naive_bayes':
                nb_support = has_rd_support_by_nb(df, het_del_fit, hom_del_fit, params, p_cnv)
                has_rd_support=nb_support
            elif method=='hybrid':
                ls_support, nb_support, hybrid_support = has_rd_support_hybrid(df, het_del_fit, hom_del_fit, params, p_cnv, slope_threshold, rsquared_threshold, num_pos_samps)
                has_rd_support=hybrid_support

            if has_rd_support:
               vcf_out.write(line)
            else:
                for m_var in to_bnd_strings(var, True):
                    vcf_out.write(m_var + '\n')

            if diag_outfile is not None:
              svlen=df['svlen'][0]
              outf.write(var.var_id+"\t"+svtype+"\t"+str(svlen)+"\t"+str(num_pos_samps)+"\t"+str(nb_support)+"\t"+str(ls_support)+"\t"+str(hybrid_support)+"\t"+str(has_rd_support)+"\n")

    vcf_out.close()
    if diag_outfile is not None:
        outf.close()
    vcf_in.close()
    vcf_out.close()
    gender_file.close()
    if exclude_file is not None:
        exclude_file.close()

    return


def get_ae_dict(ae_path):
    if ae_path.endswith('.gz'):
        ae_bedfile = gzip.open(ae_path, 'rb')
    else:
        ae_bedfile = open(ae_path, 'r')
    ae_dict = {}
    for line in ae_bedfile:
        v = line.rstrip().split('\t')
        if len(v) < 4:
            continue
        v[1] = int(v[1])
        v[2] = int(v[2])
        if v[0] in ae_dict:
            ae_dict[v[0]].append(v[1:])
        else:
            ae_dict[v[0]] = [v[1:]]
    ae_bedfile.close()
    return ae_dict

def run_reclassifier(vcf_file, vcf_out, sex_file, ae_path, f_overlap, exclude_list, slope_threshold, rsquared_threshold, training_data, method, diag_outfile):

    ae_dict = None
    params = None
    het_del_fit = None
    hom_del_fit = None
    p_cnv=0.5       # prior probability that CNV is real

    if ae_path is not None:
        sys.stderr.write("loading annotations\n")
        ae_dict=get_ae_dict(ae_path)
    
    if(method!="large_sample"):
          sys.stderr.write("calculating parameters\n")
          #calculate per-sample CN profiles on training set
          [params, het_del_fit, hom_del_fit]=calc_params(training_data)

    sys.stderr.write("reclassifying\n")
    sv_classify(vcf_file,
                vcf_out,
                sex_file,
                exclude_list,
                ae_dict,
                f_overlap,
                slope_threshold,
                rsquared_threshold,
                p_cnv,
                het_del_fit,
                hom_del_fit,
                params, 
                diag_outfile,
                method)

def add_arguments_to_parser(parser):
    parser.add_argument('-i', '--input', metavar='<VCF>', dest='vcf_in', type=argparse.FileType('r'), default=None, help='VCF input [stdin]')
    parser.add_argument('-o', '--output', metavar='<VCF>', dest='vcf_out', type=argparse.FileType('w'), default=sys.stdout, help='VCF output [stdout]')
    parser.add_argument('-g', '--gender', metavar='<FILE>', dest='gender', type=argparse.FileType('r'), required=True, default=None, help='tab delimited file of sample genders (male=1, female=2)\nex: SAMPLE_A\t2')
    parser.add_argument('-a', '--annotation', metavar='<BED>', dest='ae_path', type=str, default=None, help='BED file of annotated elements')
    parser.add_argument('-f', '--fraction', metavar='<FLOAT>', dest='f_overlap', type=float, default=0.9, help='fraction of reciprocal overlap to apply annotation to variant [0.9]')
    parser.add_argument('-e', '--exclude', metavar='<FILE>', dest='exclude', type=argparse.FileType('r'), required=False, default=None, help='list of samples to exclude from classification algorithms')
    parser.add_argument('-s', '--slope_threshold', metavar='<FLOAT>', dest='slope_threshold', type=float, default=1.0, help='minimum slope absolute value of regression line to classify as DEL or DUP[1.0]')
    parser.add_argument('-r', '--rsquared_threshold', metavar='<FLOAT>', dest='rsquared_threshold', type=float, default=0.2, help='minimum R^2 correlation value of regression line to classify as DEL or DUP [0.2], for large sample reclassification')
    parser.add_argument('-t', '--tSet', metavar='<STRING>', dest='tSet', type=str, default=None, required=False, help='high quality deletions & duplications training dataset[vcf], required by naive Bayes reclassification')
    parser.add_argument('-m', '--method', metavar='<STRING>', dest='method', type=str, default="large_sample", required=False, help='reclassification method, one of (large_sample, naive_bayes, hybrid)', choices=['large_sample', 'naive_bayes', 'hybrid'])
    parser.add_argument('-d', '--diag_file', metavar='<STRING>', dest='diag_outfile', type=str, default=None, required=False, help='text file to output method comparisons')
    parser.set_defaults(entry_point=run_from_args)

def description():
    return 'reclassify DEL and DUP based on read depth information'

def command_parser():
    parser = argparse.ArgumentParser(description=description())
    add_arguments_to_parser(parser)
    return parser

def run_from_args(args):
    if args.tSet is None:
        if args.method!="large_sample":
            sys.stderr.write("Training data required for naive Bayes or hybrid classifiers\n")
            parser.print_help()
            sys.exit(1)
    with su.InputStream(args.vcf_in) as stream:
        run_reclassifier(stream, args.vcf_out, args.gender, args.ae_path, args.f_overlap, args.exclude, args.slope_threshold, args.rsquared_threshold, args.tSet, args.method, args.diag_outfile)

if __name__ == '__main__':
    parser = command_parser()
    args=parser.parse_args()
    sys.exit(args.entry_point(args))



import svtools.l_bp as l_bp

import sys
import os
import heapq
import argparse
from tempfile import gettempdir
from collections import namedtuple

Keyed = namedtuple("Keyed", ["key", "obj"])
def merge(*iterables):
   keyed_iterables = [(Keyed(l_bp.vcf_line_key(obj), obj) for obj in iterable)
                        for iterable in iterables]
   for element in heapq.merge(*keyed_iterables):
       yield element.obj

class Lsort(object):
    def __init__(self, vcf_file_names, tempdir=None, batchsize=200):
        if tempdir:
            self.tempdir = tempdir
        else:
            self.tempdir = gettempdir()
        self.batchsize = batchsize
        self.vcf_file_names = vcf_file_names
        self.vcf_lines = []
        self.vcf_headers = []
        self.temp_files = []

    def execute(self):
        
        counter = 0
        for vcf_file_name in self.vcf_file_names:
            samples = l_bp.parse_vcf(vcf_file_name, self.vcf_lines, self.vcf_headers)
            for sample in samples:
                self.vcf_headers.append("##SAMPLE=<ID=" + sample + ">\n")
            counter += 1
            if counter > self.batchsize:
                self.vcf_lines.sort(key=l_bp.vcf_line_key)
                self.write_temp_file()
                counter = 0
        # no need to write the final batch to file
        self.write_header()

        self.vcf_lines.sort(key=l_bp.vcf_line_key)
        iterables = self.temp_files + [self.vcf_lines]
        sys.stdout.writelines(merge(*iterables))

    def close_tempfiles(self):
        for tmp in self.temp_files:
            tmp.close()
            os.remove(tmp.name)

    def write_header(self):
        self.vcf_headers.append("##INFO=<ID=SNAME,Number=.,Type=String," + \
            "Description=\"Source sample name\">\n")
        self.vcf_headers.append("##INFO=<ID=ALG,Number=1,Type=String," + \
            "Description=\"Evidence PDF aggregation algorithm\">\n")
        self.vcf_headers.append("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\t" + \
            "VARIOUS\n")
        self.vcf_headers.sort(cmp=l_bp.header_line_cmp)
        sys.stdout.writelines(self.vcf_headers)

    def write_temp_file(self):
        temp_outfile = open(os.path.join(self.tempdir,'%06i'%len(self.temp_files)),'w+b',64*1024)
        temp_outfile.writelines(self.vcf_lines)
        temp_outfile.flush()
        temp_outfile.seek(0)
        self.temp_files.append(temp_outfile)

        #vcf_line array
        self.vcf_lines = []

def description():
    return 'sort N LUMPY VCF files into a single file'

def epilog():
    return '''Specify -t to override where temporary files are placed. Use -b to control the amount of memory required. 
    This will vary depending on the number of lines in your input files.'''

def add_arguments_to_parser(parser):
    parser.add_argument('vcf_files', metavar='<VCF>', nargs='+', help='VCF files to combine and sort')
    parser.add_argument('-t', '--tempdir', metavar='<DIRECTORY_PATH>', default=gettempdir(), help='temporary directory')
    parser.add_argument('-b', '--batchsize', metavar='<INT>', type=int, default=200, help='number of files to sort in batch')
    parser.set_defaults(entry_point=run_from_args)

def command_parser():
    parser = argparse.ArgumentParser(description=description(), epilog=epilog())
    add_arguments_to_parser(parser)
    return parser

def run_from_args(args):
    sorter = Lsort(args.vcf_files, tempdir=args.tempdir, batchsize=args.batchsize)
    sorter.execute()

if __name__ == "__main__":
    parser = command_parser()
    args = parser.parse_args()
    sys.exit(args.entry_point(args))
